package de.meinefirma.meinprojekt.ftp;

import java.io.*;
import java.net.*;
import java.util.*;
import java.util.concurrent.*;
import java.util.concurrent.atomic.AtomicLong;
import java.util.regex.Pattern;
import java.util.zip.*;
import org.apache.commons.net.*;
import org.apache.commons.net.ftp.*;
import org.apache.log4j.Logger;

/**
 * <b> Wrapper fuer transparente Dateizugriffe entweder per FTP oder per Dateisystem (lokal oder Fileserver). </b><br>
 * <br>
 * Beispielsweise kann konfiguriert werden, dass Dateizugriffe in der Produktionsumgebung per FTP,
 * aber in lokalen Entwicklertestumgebungen per lokalem Dateisystem erfolgen.<br>
 * Die FTP-Methoden basieren auf Apache FTPClient:
 * {@link "http://commons.apache.org/net/apidocs/org/apache/commons/net/ftp/FTPClient.html"}
 * <br>
 * TASKS_COUNT sollte auf ca. ein Zehntel der Anzahl der zur Verfuegung stehenden dynamischen Ports (bzw. File-Handles) gesetzt werden.
 */
public final class FtpFileWrapper
{
   private static final Logger  LOGGER = Logger.getLogger( FtpFileWrapper.class );
   public  static final boolean LOG_ALL_CONSOLE = false;
   public  static final boolean PASV            = true;
   public  static final String  FS_ROOT_KEY     = "FsRoot";
   public  static final String  FTP_URI_KEY     = "FtpUri";
   public  static final String  FTP_USR_KEY     = "FtpUsr";
   public  static final String  FTP_PWD_KEY     = "FtpPwd";
   public  static final String  FTP_HOST_KEY    = "FtpHost";
   public  static final String  FTP_PORT_KEY    = "FtpPort";
   public  static final String  FTP_PATH_KEY    = "FtpPath";
   public  static final String  FTP_PROT_URI    = "ftp://";
   public  static final String  TEMP_MARK       = "$$";
   public  static final int     UNZIP_MIN_MS    = 120;
   private static final int     BUFF_SIZE       = 32768;
   private static final long    RETRY_PAUSE_MS  = TimeUnit.MINUTES.toMillis( 1 );
   private static final int     RETRY_COUNT     = 5;
   private static final int     TASKS_COUNT     = 450;
   private static final int     TASKS_WAIT_MINUTEN = 2;
   private static final int     TASKS_WAIT_REDUCE  = TASKS_WAIT_MINUTEN * 50;
   private static final long    NANOSEK_PRO_MILLISEK = 1000000;
   private static final AtomicLong  LOGIN_COUNT      = new AtomicLong();
   private static final Queue<Long> TASKS_TIME       = new ConcurrentLinkedQueue<Long>();
   private String     fsRootDir;
   private String     fsActSubDir;
   private String     ftpRootPath;
   private String     ftpActSubDir;
   private Properties ftpOderFileProps;
   private boolean    ftpOrFile;
   private FTPClient  ftpClient = null;
   private boolean    propsOk   = false;
   private boolean    loggedin  = false;

   /**
    * Privater Konstruktor.
    * @param  ftpOderFileProps Konfigurationsparameter entweder fuer FTP oder fuer Dateisystem-Root
    */
   private FtpFileWrapper( Properties ftpOderFileProps )
   {
      this.ftpOderFileProps = ftpOderFileProps;
   }

   /**
    * Erzeuge weiteren FtpFileWrapper.
    */
   public FtpFileWrapper createSecondFtpFileWrapper()
   {
      return new FtpFileWrapper( ftpOderFileProps );
   }

   /**
    * Erzeuge FtpFileWrapper mit einem Properties-Objekt.
    * @param  ftpOderFileProps Konfigurationsparameter entweder fuer FTP oder fuer Dateisystem-Root
    */
   public static FtpFileWrapper createFtpFileWrapperFromProps( Properties ftpOderFileProps )
   {
      return new FtpFileWrapper( ftpOderFileProps );
   }

   /**
    * Erzeuge FtpFileWrapper mit einer Properties-Datei (entweder im CLASSPATH oder per Pfad).
    * @param  propFile Properties-Datei mit Konfigurationsparametern entweder fuer FTP oder fuer Dateisystem-Root
    */
   public static FtpFileWrapper createFtpFileWrapperFromPropFile( String propFile )
   {
      return new FtpFileWrapper( loadPropertiesFromFileFromDirOrClasspath( propFile ) );
   }

   /**
    * Erzeuge FtpFileWrapper aus URI (Fileserver-URI oder FTP-URI).
    * @param  uri Fileserver-URI oder FTP-URI (FTP-URI muss mit "ftp://" beginnen)
    */
   public static FtpFileWrapper createFtpFileWrapperFromUri( String uri )
   {
      Properties p = new Properties();
      p.put( FTP_URI_KEY, uri );
      return new FtpFileWrapper( p );
   }

   /**
    * Erzeuge FTP-FtpFileWrapper zu FTP-Parametern.
    */
   public static FtpFileWrapper createFtpFileWrapperFromFtpParms( String host, String port, String path, String usr, String pwd )
   {
      Properties p = new Properties();
      p.put( FTP_HOST_KEY, host );
      p.put( FTP_PORT_KEY, port );
      p.put( FTP_PATH_KEY, path );
      p.put( FTP_USR_KEY,  usr );
      p.put( FTP_PWD_KEY,  pwd );
      return new FtpFileWrapper( p );
   }

   /**
    * Um nicht in zu kurzer Zeit zu viele Logins durchzufuehren und Ports zu belegen,
    * werden bei einigen Methoden Zeitstempel gespeichert und eventuell wird verzoegert.
    */
   private static void registerTasktimeAndDelay()
   {
      long actTime = (new Date()).getTime();
      if( TASKS_TIME.isEmpty() ) {
         TASKS_TIME.add( Long.valueOf( actTime ) );
         return;
      }
      Long oldestTime = TASKS_TIME.peek();
      int  n;
      while( (n = TASKS_TIME.size()) > TASKS_COUNT ) {
         oldestTime = TASKS_TIME.poll();
      }
      long diffOldestTime = actTime - oldestTime.longValue();
      long waitTime = TimeUnit.MINUTES.toMillis( TASKS_WAIT_MINUTEN ) - diffOldestTime;
      waitTime = Math.max( 0, waitTime ) * n / TASKS_COUNT / TASKS_WAIT_REDUCE;
      if( waitTime > 0 ) {
         if( waitTime > TimeUnit.MINUTES.toMillis( TASKS_WAIT_MINUTEN ) / TASKS_WAIT_REDUCE / 2 ) {
            String diffOldestTimeString = ( diffOldestTime > 1000 ) ? ((diffOldestTime / 1000) + " s") : (diffOldestTime + " ms");
            LOGGER.info( "FTP: " + ((waitTime / 100L) / 10.) + " s warten, da es in den letzten " + diffOldestTimeString + " mehr als " + n + " FTP-Tasks gab." );
         }
         try { Thread.sleep( waitTime ); } catch( InterruptedException e ) {/*ok*/}
      }
      TASKS_TIME.add( Long.valueOf( (new Date()).getTime() ) );
   }

   /**
    * Vor Dateizugriffen muss Login und hinterher Logout erfolgen.
    * @return true falls erfolgreich
    */
   public boolean login() throws IOException
   {
      registerTasktimeAndDelay();
      LOGIN_COUNT.incrementAndGet();
      if( loggedin ) { logout(); }
      ftpActSubDir = null;
      // Properties ueberpruefen:
      if( ftpOderFileProps == null || ftpOderFileProps.isEmpty() ) {
         throw new IllegalArgumentException( "Fehler: Properties fehlen." );
      }
      if( checkStringNichtLeer( ftpOderFileProps.getProperty( FTP_HOST_KEY ) ) &&
          checkStringNichtLeer( ftpOderFileProps.getProperty( FTP_PORT_KEY ) ) ) {
         propsOk   = true;
         ftpOrFile = true;
      } else if( checkStringNichtLeer( ftpOderFileProps.getProperty( FTP_URI_KEY ) ) ) {
         String uri = ftpOderFileProps.getProperty( FTP_URI_KEY );
         ftpOrFile = uri != null && uri.trim().toLowerCase( Locale.GERMAN ).startsWith( FTP_PROT_URI );
         propsOk   = true;
         if( ftpOrFile ) {
            Properties p = extractFtpParmsFromUri( uri );
            if( checkStringNichtLeer( p.getProperty( FTP_HOST_KEY ) ) &&
                checkStringNichtLeer( p.getProperty( FTP_PORT_KEY ) ) ) {
               ftpOderFileProps = p;
            } else {
               propsOk = false;
            }
         } else {
            ftpOderFileProps.setProperty( FS_ROOT_KEY, uri );
         }
      } else if( checkStringNichtLeer( ftpOderFileProps.getProperty( FS_ROOT_KEY ) ) ) {
         propsOk   = true;
         ftpOrFile = false;
      }
      if( !propsOk ) {
         throw new IllegalArgumentException( "Fehler: Properties entweder zum FTP-Zugang oder zur Fileserver-Root fehlen." );
      }
      if( ftpOrFile ) {
         // FTP:
         int port = 0;
         try {
            port = Integer.parseInt( ftpOderFileProps.getProperty( FTP_PORT_KEY ).trim() );
         } catch( NumberFormatException e ) {/* nothing to do*/}
         if( port == 0 ) {
            throw new IllegalArgumentException( "Fehler: Properties zum FTP-Zugang muessen gueltige Portnummer enthalten." );
         }
         ftpClient = new FTPClient();
         if( LOG_ALL_CONSOLE ) { ftpClient.addProtocolCommandListener( new ConsoleProtocolCommandListener() ); }
         ftpClient.connect( ftpOderFileProps.getProperty( FTP_HOST_KEY ).trim(), port );
         try {
            ftpClient.login( ftpOderFileProps.getProperty( FTP_USR_KEY ),
                             ftpOderFileProps.getProperty( FTP_PWD_KEY ) );
            loggedin = FTPReply.isPositiveCompletion( ftpClient.getReplyCode() );
            if( !loggedin ) {
               throw new IOException( "FTP-Fehler: " + Arrays.asList( ftpClient.getReplyStrings() ) );
            }
            ftpRootPath = ftpOderFileProps.getProperty( FTP_PATH_KEY );
            if( ftpRootPath != null ) { ftpRootPath = ftpRootPath.trim(); }
            if( ftpRootPath != null && ftpRootPath.length() > 0 ) {
               if( !ftpClient.changeWorkingDirectory( ftpRootPath ) ) {
                  throw new IOException( "FTP-Fehler mit FTP-Root-Path '" + ftpRootPath + "': " + Arrays.asList( ftpClient.getReplyStrings() ) );
               }
            }
            if( PASV ) { ftpClient.enterLocalPassiveMode(); }
            ftpClient.setFileTransferMode( FTP.BINARY_FILE_TYPE );
            ftpClient.setFileType( FTP.BINARY_FILE_TYPE );
         } catch( IOException ex ) {
            ftpClient.disconnect();
            ftpClient = null;
            throw ex;
         }
      } else {
         // Dateisystem:
         fsActSubDir = "";
         fsRootDir = ftpOderFileProps.getProperty( FS_ROOT_KEY );
         if( !checkStringNichtLeer( fsRootDir ) ) {
            throw new IllegalArgumentException( "Fehler: Fileserver-Root-Verzeichnis muss angegeben werden." );
         }
         fsRootDir = fsRootDir.trim();
         File fsRootDirFl = new File( fsRootDir );
         loggedin = fsRootDirFl.exists() && fsRootDirFl.isDirectory();
         if( !loggedin ) {
            loggedin = fsRootDirFl.mkdirs();
         }
      }
      return loggedin;
   }

   /**
    * Vor Dateizugriffen muss Login und hinterher Logout erfolgen.
    */
   public void logout()
   {
      loggedin = false;
      ftpActSubDir = null;
      if( ftpClient != null ) {
         try {
            try {
               ftpClient.logout();
            } finally {
               ftpClient.disconnect();
            }
         } catch( IOException ex ) {
            /* nothing to do*/
         } finally {
            ftpClient = null;
         }
      }
   }

   /**
    * FTP- oder Fileserver-Betrieb.
    * @return true falls FTP, false falls Fileserver
    */
   public boolean getFtpOrFile()
   {
      loginIfNotLoggedInISE();
      return ftpOrFile;
   }

   /**
    * Einige wenige Methoden erfordern dieses zusaetzliche abschliessende Kommando,
    * z.B. die Methoden, welche {@link FTPClient#retrieveFileStream()} verwenden
    * (insbesondere {@link FtpFileWrapper#retrieveTextFile()}, {@link FtpFileWrapper#retrieveBinaryFile()}
    *  und {@link FtpFileWrapper#storeFile( String remoteResultFile )}).<br>
    * Siehe {@link "http://commons.apache.org/net/apidocs/org/apache/commons/net/ftp/FTPClient.html#completePendingCommand()"}
    */
   public boolean completePendingCommand()
   {
      if( ftpOrFile ) {
         try {
            // if( FTPReply.isPositiveIntermediate( ftpClient.getReplyCode() ) )
            ftpClient.completePendingCommand();
         } catch( IOException ex ) {
            LOGGER.info( "completePendingCommand(): ", ex );
         }
         return FTPReply.isPositiveCompletion( ftpClient.getReplyCode() );
      }
      return true;
   }

   /**
    * Wechseln des aktuellen Verzeichnisses.
    * @param  subDir neues Verzeichnis
    * @return true falls erfolgreich
    */
   public boolean changeWorkingDirectory( String subDir ) throws IOException
   {
      loginIfNotLoggedIn();
      if( subDir == null || subDir.trim().length() == 0 ) { return true; }
      String verz = subDir.trim();
      if( ftpOrFile ) {
         verz = verz.replace( '\\', '/' );
         if( verz.startsWith( "/" ) && ftpRootPath != null ) {
            verz = ftpRootPath + verz;
            if( !verz.startsWith( "/" ) ) { verz = "/" + verz; }
         }
         if( ftpClient.changeWorkingDirectory( verz ) ) {
            ftpActSubDir = ftpClient.printWorkingDirectory();
            return true;
         }
         return false;
      }
      String fsActSubDirOld = fsActSubDir;
      if( verz.startsWith( "/" ) || verz.startsWith( "\\" ) ) {
         fsActSubDir = ( verz.length() > 1 ) ? verz : "";
      } else {
         fsActSubDir = fsActSubDir + File.separator + verz;
      }
      if( (new File( fsRootDir + fsActSubDir )).exists() ) {
         return true;
      }
      fsActSubDir = fsActSubDirOld;
      return false;
   }

   /**
    * Returniere das aktuelle Verzeichnis.
    * @return aktuelles Verzeichnis
    */
   public String getWorkingDirectory() throws IOException
   {
      loginIfNotLoggedIn();
      if( ftpOrFile ) {
         String dir = ftpClient.printWorkingDirectory();
         if( ftpRootPath != null && ftpRootPath.length() > 0 ) {
            int i = dir.indexOf( ftpRootPath );
            dir = dir.substring( i + ftpRootPath.length() );
         }
         return ( dir != null && dir.length() > 0 ) ? dir : "/";
      }
      return fsActSubDir;
   }

   /**
    * Erstellen eines neuen Verzeichnisses.
    * @param  subDir neues Verzeichnis
    * @return true falls erfolgreich
    */
   public boolean makeDirectory( String subDir ) throws IOException
   {
      loginIfNotLoggedIn();
      if( subDir == null || subDir.trim().length() == 0 ) { return false; }
      String verz = subDir.trim();
      if( ftpOrFile ) {
         verz = verz.replace( '\\', '/' );
         if( verz.endsWith( "/" ) ) { verz = verz.substring( 0, verz.length() - 1 ); }
         String[]      vv = verz.split( "/" );
         StringBuilder sb = new StringBuilder();
         boolean       b  = false;
         if( vv[0].length() == 0 && ftpRootPath != null ) {
            vv[0] = ftpRootPath;
         }
         for( int i = 0; i < vv.length; i++ ) {
            sb.append( vv[i].trim() );
            if( sb.length() > 0 ) { b = ftpClient.makeDirectory( sb.toString() ); }
            sb.append( "/" );
         }
         return b;
      }
      return (new File( getFsRootActPath( verz ) + verz )).mkdirs();
   }

   /**
    * Lesen von Verzeichnisnamen.
    * @return Liste der Verzeichnisnamen
    */
   public List<String> listDirNames() throws IOException
   {
      return listNames( null, true, true );
   }
   /**
    * Lesen von Verzeichnisnamen in einem Unterverzeichnis.
    * @param  subDir Lesen der Verzeichnisnamen in diesem Unterverzeichnis
    * @return Liste der Verzeichnisnamen
    */
   public List<String> listDirNames( String subDir ) throws IOException
   {
      return listNames( subDir, true, true );
   }
   /**
    * Lesen von Dateinamen.
    * @return Liste der Dateinamen
    */
   public List<String> listFileNames() throws IOException
   {
      return listNames( null, false, true );
   }
   /**
    * Lesen von Dateinamen in einem Unterverzeichnis.
    * @param  subDir Lesen der Dateinamen in diesem Unterverzeichnis
    * @return Liste der Dateinamen
    */
   public List<String> listFileNames( String subDir ) throws IOException
   {
      return listNames( subDir, false, true );
   }

   /**
    * Lesen entweder von Verzeichnisnamen oder von Dateinamen in einem Unterverzeichnis.
    * @param  dirOrFiles    true falls Verzeichnisnamen gelesen werden sollen oder false fuer Dateinamen
    * @param  subDir        Unterverzeichnis
    * @param  ohneTempFiles true falls auf "$$" (= TEMP_MARK) endende Dateien nicht beruecksichtigt werden sollen
    * @return Liste der Verzeichnisnamen bzw. Dateinamen
    */
   private List<String> listNames( String subDir, final boolean dirOrFiles, final boolean ohneTempFiles ) throws IOException
   {
      loginIfNotLoggedIn();
      List<String> namesList = new ArrayList<String>();
      if( ftpOrFile ) {
         // FTP:
         FTPFile[] files = listFtpFiles( subDir, false );
         if( files != null ) {
            for( FTPFile f : files ) {
               if( f != null && ((dirOrFiles && f.isDirectory()) || (!dirOrFiles && f.isFile()))
                   && (!ohneTempFiles || !f.getName().endsWith( TEMP_MARK )) ) {
                      namesList.add( f.getName() );
               }
            }
         }
         return namesList;
      }
      // Dateisystem:
      File[] files = listFsFiles( subDir );
      if( files != null ) {
         for( File f : files ) {
            if( f != null && ((dirOrFiles && f.isDirectory()) || (!dirOrFiles && f.isFile()))
                && (!ohneTempFiles || !f.getName().endsWith( TEMP_MARK )) ) {
                   namesList.add( f.getName() );
            }
         }
      }
      return namesList;
   }

   /**
    * Ermittle die Groesse einer Remote-Datei.
    * @param  file Remote-Datei
    * @return Groesse der Remote-Datei (in Bytes)
    */
   public long getLength( String file )
   {
      if( file == null || file.trim().length() == 0 ) {
         return 0;
      }
      loginIfNotLoggedInISE();
      if( ftpOrFile ) {
         // FTP:
         File   fl = new File( file.trim() );
         String parentPath = fl.getParent();
         String fileName   = fl.getName();
         if( fileName == null || fileName.length() == 0 ) { return 0; }
         try {
            FTPFile[] files = listFtpFiles( parentPath, false );
            if( files != null ) {
               for( FTPFile f : files ) {
                  if( f != null && fileName.equals( f.getName() ) ) {
                     return f.getSize();
                  }
               }
            }
         } catch( IOException ex ) {/* nothing to do */}
         return 0;
      }
      // Dateisystem:
      return (new File( getFsRootActPath( file ) + file )).length();
   }

   /**
    * Ermittle Anzahl der Remote-Dateien und Summe der Dateigroessen der Remote-Dateien.
    * @param  subDir     Unterverzeichnis
    * @param  ignoreDirs zu ignorierende Unterverzeichnisse
    * @return Zwei Werte: a) Anzahl Dateien, b) Summe der Dateigroessen in Byte
    */
   public long[] getFileCountAndSizeSum( String subDir, Set<String> ignoreDirs )
   {
      loginIfNotLoggedInISE();
      long[] countAndSizeSum = new long[2];
      List<FtpFsFileData> fds = listFtpFsFileData( subDir, false );
      if( fds == null ) { return null; }
      for( FtpFsFileData f : fds ) {
         if( f.isFile ) {
            countAndSizeSum[0] = countAndSizeSum[0] + 1;
            countAndSizeSum[1] = countAndSizeSum[1] + f.size;
         } else if( f.isDirectory ) {
            if( ignoreDirs == null || !ignoreDirs.contains( f.name ) ) {
               String dir = ( subDir == null || subDir.trim().length() == 0 ) ? "" : subDir + "/";
               long[] countAndSizeSumDir = getFileCountAndSizeSum( dir + f.name, ignoreDirs );
               countAndSizeSum[0] += countAndSizeSumDir[0];
               countAndSizeSum[1] += countAndSizeSumDir[1];
            }
         }
      }
      return countAndSizeSum;
   }

   /**
    * Ermittle den Zeitstempel der aeltesten Remote-Datei.
    * @param  startDir                 Unterverzeichnis
    * @param  regexAllowedCompleteDirs Regex-Ausdruck fuer zu verwendende komplette Unterverzeichnispfade (also inkl. aller Unterverzeichnisse)
    * @param  ignoreSingleDirs         zu ignorierende Unterverzeichnisse, aber nicht komplette Unterverzeichnispfade, sondern einzelne Unterverzeichnisse
    * @return FileData zur aeltesten Remote-Datei
    */
   public FtpFsFileData getOldestFileData( String startDir, String regexAllowedCompleteDirs, Set<String> ignoreSingleDirs )
   {
      loginIfNotLoggedInISE();
      List<FtpFsFileData> fds = listFtpFsFileData( startDir, false );
      if( fds == null ) { return null; }
      String prefixDir = ( startDir == null || startDir.trim().length() == 0 ) ? "" : (( startDir.trim().endsWith( "/" ) ) ? startDir.trim() : startDir.trim() + "/");
      Pattern regexAllowedCompleteDirsPattern = ( regexAllowedCompleteDirs != null && regexAllowedCompleteDirs.trim().length() > 0 ) ? Pattern.compile( regexAllowedCompleteDirs.trim() ) : null;
      boolean noRootFiles = ignoreSingleDirs != null && ignoreSingleDirs.contains( "/" );
      FtpFsFileData fdOldest = null;
      for( FtpFsFileData fd : fds ) {
         if( fd.isFile ) {
            if( (!noRootFiles || !"/".equals( prefixDir )) &&
                  (fdOldest == null || fdOldest.timestamp.after( fd.timestamp )) ) {
               fdOldest = fd;
            }
         } else if( fd.isDirectory ) {
            if( (regexAllowedCompleteDirsPattern == null || regexAllowedCompleteDirsPattern.matcher( prefixDir + fd.name ).matches()) &&
                  (ignoreSingleDirs == null || !ignoreSingleDirs.contains( fd.name )) ) {
               FtpFsFileData fdAct = getOldestFileData( prefixDir + fd.name, regexAllowedCompleteDirs, ignoreSingleDirs );
               if( fdAct != null && (fdOldest == null || fdOldest.timestamp.after( fdAct.timestamp )) ) {
                  fdOldest = fdAct;
               }
            }
         }
      }
      return fdOldest;
   }

   /**
    * Returniert ob eine Remote-Datei existiert.
    * @param  file Remote-Datei
    * @return true falls Remote-Datei existiert
    */
   public boolean existsFile( String file )
   {
      loginIfNotLoggedInISE();
      if( file == null || file.trim().length() == 0 ) { return false; }
      if( ftpOrFile ) {
         // FTP:
         File   f = new File( file.trim() );
         String parentPath = f.getParent();
         String fileName   = f.getName();
         try {
            List<String> filenames = listNames( parentPath, false, false );
            return filenames != null && filenames.contains( fileName );
         } catch (IOException e) {
            return false;
         }
      }
      // Dateisystem:
      File f = new File( getFsRootActPath( file ) + file );
      return f.exists() && f.isFile();
   }

   /**
    * Returniert ob eine Remote-Datei existiert (versucht es mehrmals).
    * @param  file Remote-Datei
    * @return true falls Remote-Datei existiert
    */
   public boolean existsFileWithLoop( String file )
   {
      boolean b = false;
      int i = 0;
      while( !b && i++ <= RETRY_COUNT ) {
         b = existsFile( file );
         if( !b ) { try { Thread.sleep( RETRY_PAUSE_MS ); } catch( InterruptedException e ) {/*ok*/} }
      }
      if( b && i > 1 ) {
         LOGGER.warn( "FTP-Warnung: Datei '" + file + "' erst beim " + i + ". Versuch gefunden." );
      }
      return b;
   }

   /**
    * Returniert ob ein Remote-Verzeichnis existiert.
    * @param  dir  Remote-Verzeichnis
    * @return true falls Remote-Verzeichnis existiert
    */
   public boolean existsDir( String dir )
   {
      loginIfNotLoggedInISE();
      if( dir == null || dir.trim().length() == 0 ) { return false; }
      if( ftpOrFile ) {
         // FTP:
         File   f = new File( dir.trim() );
         String parentPath = f.getParent();
         String dirName    = f.getName();
         try {
            List<String> dirnames = listNames( parentPath, true, false );
            return dirnames != null && dirnames.contains( dirName );
         } catch (IOException e) {
            return false;
         }
      }
      // Dateisystem:
      File f = new File( getFsRootActPath( dir ) + dir );
      return f.exists() && f.isDirectory();
   }

   /**
    * Loesche Datei.
    * @param  file zu entfernende Remote-Datei
    * @return true falls erfolgreich
    */
   public boolean deleteFile( String file ) throws IOException
   {
      loginIfNotLoggedIn();
      if( file == null || file.trim().length() == 0 ) { return false; }
      if( ftpOrFile ) {
         String fl = file.trim().replace( '\\', '/' );
         if( fl.startsWith( "/" ) && ftpRootPath != null && ftpRootPath.length() > 0 ) {
            fl = ftpRootPath + fl;
         }
         return ftpClient.deleteFile( fl );
      }
      return (new File( getFsRootActPath( file ) + file )).delete();
   }

   /**
    * Loesche Verzeichnis inklusive aller Unterverzeichnisse und aller enthaltenen Dateien.
    * @param  dir  zu loeschendes Remote-Verzeichnis
    * @return true falls erfolgreich
    */
   public boolean removeDirectory( String dir ) throws IOException
   {
      loginIfNotLoggedIn();
      if( dir == null || dir.trim().length() == 0 ) { return false; }
      boolean b = true;
      List<FtpFsFileData> fds = listFtpFsFileData( dir, false );
      if( fds == null ) { return false; }
      for( FtpFsFileData f : fds ) {
         String fn = dir + File.separator + f.name;
         if( f.isFile ) {
            b &= ( ftpOrFile )
                 ? ftpClient.deleteFile( fn.replace( '\\', '/' ) )
                 : (new File( getFsRootActPath( fn ) + fn )).delete();
         } else if( f.isDirectory ) {
            b &= removeDirectory( fn );
         }
      }
      b &= ( ftpOrFile )
           ? ftpClient.removeDirectory( dir.replace( '\\', '/' ) )
           : (new File( getFsRootActPath( dir ) + dir )).delete();
      return b;
   }

   /**
    * Umbenenne Datei.
    * @param oldFilename Vorheriger Dateiname
    * @param newFilename Neuer Dateiname
    * @return true falls erfolgreich
    */
   public boolean renameFile( String oldFilename, String newFilename ) throws IOException
   {
      loginIfNotLoggedIn();
      if( oldFilename == null || oldFilename.trim().length() == 0 ||
          newFilename == null || newFilename.trim().length() == 0 ||
          oldFilename.trim().equals( newFilename.trim() ) ) {
         return false;
      }
      if( ftpOrFile ) {
         String oldFn = oldFilename.trim().replace( '\\', '/' );
         String newFn = newFilename.trim().replace( '\\', '/' );
         if( oldFn.startsWith( "/" ) && ftpRootPath != null ) { oldFn = ftpRootPath + oldFn; }
         if( newFn.startsWith( "/" ) && ftpRootPath != null ) { newFn = ftpRootPath + newFn; }
         return ftpClient.rename( oldFn, newFn );
      }
      return (new File( getFsRootActPath( oldFilename ) + oldFilename.trim() )).renameTo(
              new File( getFsRootActPath( newFilename ) + newFilename.trim() ) );
   }

   /**
    * Kopiere Remote-Datei in andere Remote-Datei (bei beiden Dateien kann auch ein Unterverzeichnis angegeben werden).
    * @param  remoteSourceFile Remote-Quell-Datei
    * @param  remoteResultFile Remote-Ergebnis-Datei
    * @return true falls erfolgreich
    */
   public boolean copyRemoteFile( String remoteSourceFile, String remoteResultFile ) throws IOException
   {
      if( remoteSourceFile == null || remoteResultFile == null ) { throw new IllegalArgumentException(); }
      if( remoteSourceFile.equals( remoteResultFile ) ) { return false; }
      FtpFileWrapper ffwrapSource = new FtpFileWrapper( ftpOderFileProps );
      FtpFileWrapper ffwrapResult = new FtpFileWrapper( ftpOderFileProps );
      ffwrapSource.login();
      try {
         ffwrapResult.login();
         try {
            String wd = getWorkingDirectory();
            if( wd != null && wd.trim().length() > 0 && !"/".equals( wd ) ) {
               ffwrapSource.changeWorkingDirectory( wd );
               ffwrapResult.changeWorkingDirectory( wd );
            }
            return copyRemoteFile( ffwrapSource, remoteSourceFile, ffwrapResult, remoteResultFile );
         } finally {
            ffwrapResult.logout();
         }
      } finally {
         ffwrapSource.logout();
      }
   }

   /**
    * Kopiere Remote-Datei vom Source-FtpFileWrapper in andere Remote-Datei vom Result-FtpFileWrapper
    * (bei beiden Dateien kann auch ein Unterverzeichnis angegeben werden).<br>
    * Zu den beiden FtpFileWrapper erfolgt kein logout(), dieser muss extern erfolgen.<br>
    * Anders <tt>copyRemoteFileWithNoRetry()</tt> fuehrt diese Methode bei Fehlschlag mehrere Versuche aus.
    * @param  ffwrapSource     FtpFileWrapper fuer Remote-Quell-Datei
    * @param  remoteSourceFile Remote-Quell-Datei
    * @param  ffwrapResult     FtpFileWrapper fuer Remote-Ergebnis-Datei
    * @param  remoteResultFile Remote-Ergebnis-Datei
    * @return true falls erfolgreich
    */
   public static boolean copyRemoteFile( FtpFileWrapper ffwrapSource, String remoteSourceFile, FtpFileWrapper ffwrapResult, String remoteResultFile ) throws IOException
   {
      int i = 0;
      while( true ) {
         try {
            copyRemoteFileWithNoRetry( ffwrapSource, remoteSourceFile, ffwrapResult, remoteResultFile );
            if( i > 0 ) {
               LOGGER.warn( "FTP-Warnung: Copy von '" + remoteSourceFile + "' nach '" + remoteResultFile + "' erst nach " + i + " Minuten erfolgreich." );
            }
            return true;
         } catch( Exception ex ) {
            if( i++ >= RETRY_COUNT ) {
               throw new IOException( "Fehler: Copy von '" + remoteSourceFile + "' nach '" + remoteResultFile + "' fehlgeschlagen. ", ex );
            }
            try { Thread.sleep( RETRY_PAUSE_MS ); } catch( InterruptedException e ) {/* ok */}
         }
      }
   }

   /**
    * Kopiere Remote-Datei vom Source-FtpFileWrapper in andere Remote-Datei vom Result-FtpFileWrapper
    * (bei beiden Dateien kann auch ein Unterverzeichnis angegeben werden).<br>
    * Zu den beiden FtpFileWrapper erfolgt kein logout(), dieser muss extern erfolgen.<br>
    * Anders <tt>copyRemoteFile()</tt> fuehrt diese Methode bei Fehlschlag nicht mehrere Versuche aus.
    * @param  ffwrapSource     FtpFileWrapper fuer Remote-Quell-Datei
    * @param  remoteSourceFile Remote-Quell-Datei
    * @param  ffwrapResult     FtpFileWrapper fuer Remote-Ergebnis-Datei
    * @param  remoteResultFile Remote-Ergebnis-Datei
    */
   public static void copyRemoteFileWithNoRetry( FtpFileWrapper ffwrapSource, String remoteSourceFile, FtpFileWrapper ffwrapResult, String remoteResultFile ) throws IOException
   {
      if( ffwrapSource == null || remoteSourceFile == null || ffwrapResult == null || remoteResultFile == null ) { throw new IllegalArgumentException(); }
      ffwrapSource.loginIfNotLoggedIn();
      ffwrapResult.loginIfNotLoggedIn();
      String remoteResultFileTmp = remoteResultFile + TEMP_MARK;
      // Existiert Quelldatei?
      if( !ffwrapSource.existsFile( remoteSourceFile ) ) {
         throw new IllegalArgumentException( "Fehler: Zu kopierende Datei '" + remoteSourceFile + "' existiert nicht." );
      }
      // Loesche beide Zieldateien falls vorhanden:
      if( ffwrapResult.existsFile( remoteResultFile ) && !ffwrapResult.deleteFile( remoteResultFile ) ) {
         throw new IllegalArgumentException( "Fehler: '" + remoteResultFile + "' existiert bereits und kann nicht geloescht werden." );
      }
      if( ffwrapResult.existsFile( remoteResultFileTmp ) && !ffwrapResult.deleteFile( remoteResultFileTmp ) ) {
         throw new IllegalArgumentException( "Fehler: '" + remoteResultFileTmp + "' existiert bereits und kann nicht geloescht werden." );
      }
      // Kopiere in Temp-Datei:
      Exception exception = null;
      boolean ret = false;
      long lenSource1 = ffwrapSource.getLength( remoteSourceFile );
      @SuppressWarnings("resource")
      InputStream is = ffwrapSource.retrieveBinaryFile( remoteSourceFile );
      if( is == null ) {
         throw new IOException( "Fehler: Copy von '" + remoteSourceFile + "' fehlgeschlagen: InputStream ist null." );
      }
      try {
         ret = ffwrapResult.storeFile( is, remoteResultFileTmp );
      } catch( Exception ex ) {
         exception = ex;
      } finally {
         is.close();
         ffwrapSource.completePendingCommand();
      }
      // Pruefe Dateigroessen:
      long lenSource2 = 0;
      long lenResult1 = 0;
      if( exception == null && ret ) {
         lenSource2 = ffwrapSource.getLength( remoteSourceFile );
         lenResult1 = ffwrapResult.getLength( remoteResultFileTmp );
      }
      if( exception != null || !ret || lenSource1 != lenSource2 || lenSource1 != lenResult1 ) {
         try { if( ffwrapResult.existsFile( remoteResultFileTmp ) ) { ffwrapResult.deleteFile( remoteResultFileTmp ); } } catch( IOException e ) {/*ok*/}
      }
      if( exception != null ) {
         throw new IOException( "Fehler: Copy von '" + remoteSourceFile + "' nach '" + remoteResultFileTmp + "' fehlgeschlagen. ", exception );
      }
      if( !ret ) {
         throw new IOException( "Fehler: Copy von '" + remoteSourceFile + "' nach '" + remoteResultFileTmp + "' fehlgeschlagen." );
      }
      if( lenSource1 != lenSource2 ) {
         throw new IOException( "Fehler: Copy fehlgeschlagen: Die Dateigroesse der Sourcedatei '" + remoteSourceFile +
               "' hat sich waehrend des Kopierens geaendert von " + lenSource1 + " Byte zu " + lenSource2 + " Byte." );
      }
      if( lenSource1 != lenResult1 ) {
         throw new IOException( "Fehler: Copy fehlgeschlagen: Dateigroesse (" + lenResult1 + " Byte) von resultierender Datei '" +
               remoteResultFileTmp + "' anders als von Sourcedatei '" + remoteSourceFile + "' (" + lenSource1 + " Byte)." );
      }
      // Rename von Temp- zu Zieldatei:
      try {
         ret &= ffwrapResult.renameFile( remoteResultFileTmp, remoteResultFile );
      } catch( Exception ex ) {
         exception = ex;
      }
      long lenResult2 = 0;
      if( exception == null && ret ) {
         lenResult2 = ffwrapResult.getLength( remoteResultFile );
      }
      if( exception != null || !ret || lenSource1 != lenResult2 ) {
         try { if( ffwrapResult.existsFile( remoteResultFile ) ) { ffwrapResult.deleteFile( remoteResultFile ); } } catch( IOException e ) {/*ok*/}
      }
      if( exception != null ) {
         throw new IOException( "Fehler: Copy/Rename von '" + remoteResultFileTmp + "' nach '" + remoteResultFile + "' fehlgeschlagen. ", exception );
      }
      if( !ret ) {
         throw new IOException( "Fehler: Copy/Rename von '" + remoteResultFileTmp + "' nach '" + remoteResultFile + "' fehlgeschlagen." );
      }
      if( lenSource1 != lenResult2 ) {
         throw new IOException( "Fehler: Copy/Rename fehlgeschlagen: Dateigroesse (" + lenResult2 + " Byte) von resultierender Datei '" +
                                remoteResultFile + "' anders als von Sourcedatei '" + remoteSourceFile + "' (" + lenSource1 + " Byte)." );
      }
   }

   /**
    * Verschiebe Remote-Datei.
    * @param  remoteSourceFile Remote-Quell-Datei
    * @param  remoteResultFile Remote-Ergebnis-Datei
    * @return true falls erfolgreich
    */
   public boolean moveRemoteFile( String remoteSourceFile, String remoteResultFile ) throws IOException
   {
      if( remoteSourceFile == null || remoteResultFile == null ) { throw new IllegalArgumentException(); }
      if( remoteSourceFile.equals( remoteResultFile ) ) { return false; }
      if( copyRemoteFile( remoteSourceFile, remoteResultFile ) ) {
         return deleteFile( remoteSourceFile );
      }
      return false;
   }

   /**
    * Verschiebe Remote-Datei.
    * Zu den beiden FtpFileWrapper erfolgt kein logout(), dieser muss extern erfolgen.
    * @param  ffwrapSource     FtpFileWrapper fuer Remote-Quell-Datei
    * @param  remoteSourceFile Remote-Quell-Datei
    * @param  ffwrapResult     FtpFileWrapper fuer Remote-Ergebnis-Datei
    * @param  remoteResultFile Remote-Ergebnis-Datei
    * @return true falls erfolgreich
    */
   public static boolean moveRemoteFile( FtpFileWrapper ffwrapSource, String remoteSourceFile, FtpFileWrapper ffwrapResult, String remoteResultFile ) throws IOException
   {
      if( remoteSourceFile == null || remoteResultFile == null || ffwrapSource == ffwrapResult ) { throw new IllegalArgumentException(); }
      if( copyRemoteFile( ffwrapSource, remoteSourceFile, ffwrapResult, remoteResultFile ) ) {
         return ffwrapSource.deleteFile( remoteSourceFile );
      }
      return false;
   }

   /**
    * Verschiebe Remote-Dateien.
    * @param  remoteSourceDir         Remote-Quell-Verzeichnis
    * @param  remoteResultDir         Remote-Ergebnis-Verzeichnis
    * @param  filenames               Liste der zu verschiebenen Dateien
    * @param  minZeitProDateiMillisek Minimale Zeit in Millisekunden pro verschobener Datei, z.B. 120 ms.
    */
   public void moveRemoteFiles( String remoteSourceDir, String remoteResultDir, List<String> filenames, int minZeitProDateiMillisek ) throws IOException
   {
      if( filenames == null || remoteSourceDir == null || remoteResultDir == null || remoteSourceDir.equals( remoteResultDir ) ) { throw new IllegalArgumentException(); }
      if( filenames.isEmpty() ) { return; }
      FtpFileWrapper ffwrapSource = new FtpFileWrapper( ftpOderFileProps );
      FtpFileWrapper ffwrapResult = new FtpFileWrapper( ftpOderFileProps );
      ffwrapSource.login();
      try {
         ffwrapResult.login();
         try {
            String wd = getWorkingDirectory();
            if( wd != null && wd.trim().length() > 0 && !"/".equals( wd ) ) {
               ffwrapSource.changeWorkingDirectory( wd );
               ffwrapResult.changeWorkingDirectory( wd );
            }
            for( String filename : filenames ) {
               boolean   ok = false;
               Exception ex = null;
               long startZeitNanosek = System.nanoTime();
               try {
                  ok = moveRemoteFile( ffwrapSource, remoteSourceDir + "/" + filename, ffwrapResult, remoteResultDir + "/" + filename );
               } catch( Exception e ) {
                  ex = e;
               }
               if( !ok || ex != null ) {
                  throw new IOException( "Fehler: '" + filename + "' konnte nicht von '" + remoteSourceDir + "' nach '" + remoteResultDir + "' verschoben werden.", ex );
               }
               long verbrauchteZeitMillisek = (System.nanoTime() - startZeitNanosek) / NANOSEK_PRO_MILLISEK;
               if( minZeitProDateiMillisek > verbrauchteZeitMillisek ) {
                  try { Thread.sleep( minZeitProDateiMillisek - verbrauchteZeitMillisek ); } catch( InterruptedException e ) {/*ok*/}
               }
            }
         } finally {
            ffwrapResult.logout();
         }
      } finally {
         ffwrapSource.logout();
      }
   }

   /**
    * Lies die ersten Textzeilen aus einer Textdatei in eine Liste von Textzeilen-Strings.<br>
    * Es wird zeilenweise gelesen, bis entweder das Dateiende erreicht ist,
    * oder die angegebene Maximalgroesse ueberschritten ist.
    * @param  remoteSourceFile Textdatei, aus der gelesen wird
    * @param  encoding         Encoding, z.B. ISO-8859-1 oder UTF-8
    * @param  maxLength        Die Anzahl der gelesenen Zeichen ueberschreitet diese Grenze um hoechstens eine Textzeile
    * @return Liste von Textzeilen-Strings
    */
   public List<String> retrieveTextLines( String remoteSourceFile, String encoding, int maxLength ) throws IOException
   {
      List<String> ss = new ArrayList<String>();
      @SuppressWarnings("resource")
      BufferedReader in = retrieveTextFile( remoteSourceFile, encoding );
      if( in == null ) { return ss; }
      try {
         String line;
         int    len = 0;
         while( len < maxLength && (line = in.readLine()) != null ) {
            len += line.length();
            ss.add( line );
         }
      } finally {
         in.close();
         completePendingCommand();
      }
      return ss;
   }

   /**
    * Zeilenweises Lesen aus einer Textdatei.<br>
    * Wichtig: Auf dem returnierten BufferedReader muss die close()-Methode aufgerufen werden
    *          und anschliessend muss completePendingCommand() aufgerufen werden!
    *          Solange der returnierte BufferedReader noch nicht geschlossen wurde,
    *          darf das Verzeichnis nicht gewechselt werden!
    * @param  remoteSourceFile Textdatei, aus der gelesen wird
    * @param  encoding         Encoding, z.B. ISO-8859-1 oder UTF-8
    * @return BufferedReader, ueber den die Textdatei zeilenweise gelesen werden kann
    */
   public BufferedReader retrieveTextFile( String remoteSourceFile, String encoding ) throws IOException
   {
      InputStream is = retrieveBinaryFile( remoteSourceFile );
      return ( is == null ) ? null : new BufferedReader( new InputStreamReader( is, encoding ) );
   }

   /**
    * Binaeres Lesen aus einer Datei.<br>
    * Wichtig: Auf dem returnierten InputStream muss die close()-Methode aufgerufen werden
    *          und anschliessend muss completePendingCommand() aufgerufen werden!
    *          Solange der returnierte InputStream noch nicht geschlossen wurde,
    *          darf das Verzeichnis nicht gewechselt werden!
    * @param  remoteSourceFile Datei, aus der gelesen wird
    * @return InputStream ueber den die Datei gelesen werden kann
    */
   public InputStream retrieveBinaryFile( String remoteSourceFile ) throws IOException
   {
      registerTasktimeAndDelay();
      loginIfNotLoggedIn();
      if( remoteSourceFile == null || remoteSourceFile.trim().length() == 0 ) { throw new IllegalArgumentException(); }
      if( ftpOrFile ) {
         // FTP:
         String fl = remoteSourceFile.trim().replace( '\\', '/' );
         if( !existsFileWithLoop( fl ) ) {
            LOGGER.error( "FTP-Fehler: '" + fl + "' existiert nicht." );
            throw new FileNotFoundException( "Fehler: '" + fl + "' existiert nicht." );
         }
         return ftpClient.retrieveFileStream( fl );
      }
      // Dateisystem:
      return new FileInputStream( getFsRootActPath( remoteSourceFile ) + remoteSourceFile.trim() );
   }

   /**
    * Speichere InputStream in Remote-Datei.<br>
    * Wichtig: Der InputStream wird nicht innerhalb der Methode geschlossen und muss ausserhalb geschlossen werden.
    * @param  is               InputStream aus dem die Daten gelesen werden
    * @param  remoteResultFile Remote-Datei in die geschrieben wird
    * @return true falls erfolgreich
    */
   public boolean storeFile( InputStream is, String remoteResultFile ) throws IOException
   {
      return storeFile( is, remoteResultFile, false );
   }

   /**
    * Speichere InputStream in Remote-Datei.<br>
    * Wichtig: Der InputStream wird nicht innerhalb der Methode geschlossen und muss ausserhalb geschlossen werden.
    * @param  is               InputStream aus dem die Daten gelesen werden
    * @param  remoteResultFile Remote-Datei in die geschrieben wird
    * @param  append           Anhaengen an existierende Datei falls true
    * @return true falls erfolgreich
    */
   public boolean storeFile( InputStream is, String remoteResultFile, boolean append ) throws IOException
   {
      registerTasktimeAndDelay();
      loginIfNotLoggedIn();
      if( is == null || remoteResultFile == null || remoteResultFile.trim().length() == 0 ) {
         throw new IllegalArgumentException();
      }
      if( !append && existsFile( remoteResultFile ) && !deleteFile( remoteResultFile ) ) {
         throw new IllegalArgumentException( "Fehler: '" + remoteResultFile + "' existiert bereits und kann nicht geloescht werden." );
      }
      if( ftpOrFile ) {
         // FTP:
         String fl = remoteResultFile.trim().replace( '\\', '/' );
         String parentPath = (new File( fl )).getParent();
         if( parentPath != null && parentPath.length() > 0 &&
               !existsDir( parentPath ) && !makeDirectory( parentPath ) ) {
            String s = ftpClient.getReplyString();
            // "!existsDir( parentPath )"-Abfrage doppelt, falls anderer Thread das Verzeichnis angelegt hat:
            if( !existsDir( parentPath ) ) {
               throw new IOException( "Fehler: Verzeichnis '" + parentPath + "' kann nicht erstellt werden: " + s );
            }
         }
         if( fl.startsWith( "/" ) && ftpRootPath != null && ftpRootPath.length() > 0 ) {
            fl = ftpRootPath + fl;
         }
         boolean ret = false;
         if( append ) {
            ret = ftpClient.appendFile( fl, is );
         } else if( ftpClient.storeFile( fl + TEMP_MARK, is ) ) {
            ret = ftpClient.rename( fl + TEMP_MARK, fl );
         }
         if( ret ) { return ret; }
         throw new IOException( "FTP-storeFile()-Fehler: " + Arrays.asList( ftpClient.getReplyStrings() ) );
      }
      // Dateisystem:
      File   f = new File( remoteResultFile.trim() );
      String parentPath = f.getParent();
      String fileName   = f.getName();
      String completeParentPath = getFsRootActPath( parentPath ) + (( parentPath != null ) ? parentPath : "");
      if( parentPath != null && parentPath.length() > 0 ) {
         File pp = new File( completeParentPath );
         if( !pp.exists() && !pp.mkdirs() ) {
            // "!pp.exists()"-Abfrage doppelt, falls anderer Thread das Verzeichnis angelegt hat:
            if( !pp.exists() ) {
               return false;
            }
         }
      }
      byte[] buff = new byte[BUFF_SIZE];
      int len;
      String fileNameTemp = ( append ) ? fileName : fileName + TEMP_MARK;
      try( FileOutputStream fos = new FileOutputStream( new File( completeParentPath, fileNameTemp ), append ) ) {
         while( 0 < (len = is.read( buff )) ) {
            fos.write( buff, 0, len );
         }
      }
      if( append ) { return true; }
      return (new File( completeParentPath, fileNameTemp )).renameTo( new File( completeParentPath, fileName ) );
   }

   /**
    * Speichere ueber returnierten OutputStream in Remote-Datei.<br>
    * Wichtig: Auf dem returnierten OutputStream muss die close()-Methode aufgerufen werden
    *          und anschliessend muss completePendingCommand() aufgerufen werden!
    *          Solange der returnierte OutputStream noch nicht geschlossen wurde,
    *          darf das Verzeichnis nicht gewechselt werden!<br>
    *          Ausserdem sollte ueberlegt werden, zuerst in einen temporaeren Dateinamen zu speichern
    *          und erst nach vollendetem Speichern ein Rename zum endgueltigen Dateinamen durchzufuehren,
    *          um das Risiko von durch andere Prozesse nur teilweise kopierten Dateien zu reduzieren.
    * @param  remoteResultFile Remote-Datei in die geschrieben wird
    * @return der OutputStream in den geschrieben wird
    */
   public OutputStream storeFile( String remoteResultFile ) throws IOException
   {
      return storeFile( remoteResultFile, false );
   }

   /**
    * Speichere ueber returnierten OutputStream in Remote-Datei.<br>
    * Wichtig: Auf dem returnierten OutputStream muss die close()-Methode aufgerufen werden
    *          und anschliessend muss completePendingCommand() aufgerufen werden!
    *          Solange der returnierte OutputStream noch nicht geschlossen wurde,
    *          darf das Verzeichnis nicht gewechselt werden!<br>
    *          Ausserdem sollte ueberlegt werden, zuerst in einen temporaeren Dateinamen zu speichern
    *          und erst nach vollendetem Speichern ein Rename zum endgueltigen Dateinamen durchzufuehren,
    *          um das Risiko von durch andere Prozesse nur teilweise kopierten Dateien zu reduzieren.
    * @param  remoteResultFile Remote-Datei in die geschrieben wird
    * @param  append           Anhaengen an existierende Datei falls true
    * @return der OutputStream in den geschrieben wird
    */
   @SuppressWarnings("resource")
   public OutputStream storeFile( String remoteResultFile, boolean append ) throws IOException
   {
      registerTasktimeAndDelay();
      loginIfNotLoggedIn();
      if( remoteResultFile == null || remoteResultFile.trim().length() == 0 ) { throw new IllegalArgumentException(); }
      if( remoteResultFile.contains( "/" ) || remoteResultFile.contains( "\\" ) ) {
         throw new IllegalArgumentException( "Kein Unterverzeichnis erlaubt in '" + remoteResultFile + "'." );
      }
      if( ftpOrFile ) {
         String fl = remoteResultFile.trim().replace( '\\', '/' );
         String parentPath = (new File( fl )).getParent();
         if( parentPath != null && parentPath.length() > 0 &&
               !existsDir( parentPath ) && !makeDirectory( parentPath ) ) {
            String s = ftpClient.getReplyString();
            // "!existsDir( parentPath )"-Abfrage doppelt, falls anderer Thread das Verzeichnis angelegt hat:
            if( !existsDir( parentPath ) ) {
               throw new IOException( "Fehler: Verzeichnis '" + parentPath + "' kann nicht erstellt werden: " + s );
            }
         }
         OutputStream os = null;
         if( append ) {
            os = ftpClient.appendFileStream( fl );
         } else {
            os = ftpClient.storeFileStream( fl );
         }
         if( os == null ) {
            String dir = getWorkingDirectory();
            dir = ( dir == null || dir.length() == 0 ) ? "" : ( dir.endsWith( "/" ) ) ? dir : (dir + "/");
            throw new IOException( "Fehler: In die Datei '" + dir + remoteResultFile + "' kann nicht geschrieben werden. " );
         }
         return os;
      }
      return new FileOutputStream( new File( getFsRootActPath( remoteResultFile ), remoteResultFile ), append );
   }

   /**
    * Konvertiere die Separator-Zeichen in einem Pfad-String zu zum jeweiligen System passenden Zeichen.
    * @param  path Input-Pfad-String
    * @return Ergebnis-Pfad-String
    */
   public String convertPathToSystemDependendPath( String path )
   {
      if( path == null ) { return null; }
      if( ftpOrFile ) { return path.replace( '\\', '/' ); }
      return path.replace( ( File.separatorChar == '\\' ) ? '/' : '\\', File.separatorChar );
   }

   /**
    * Properties aus Properties-Datei in Properties-Objekt laden.<br>
    * Die Properties-Datei kann entweder im CLASSPATH liegen oder per Pfad erreicht werden.
    * @param  propFile Dateiname der Properties-Datei
    * @return Properties-Objekt
    */
   public static Properties loadPropertiesFromFileFromDirOrClasspath( String propFile )
   {
      Properties props = new Properties();
      if( propFile == null || propFile.trim().length() == 0 ) { return props; }

      // Properties aus Datei vom CLASSPATH einlesen:
      try {
         String propFileCp = ( propFile.startsWith( "/" ) || propFile.startsWith( "\\" ) )
                             ? propFile : (File.separator + propFile);
         @SuppressWarnings("resource")
         InputStream is = ClassLoader.getSystemResourceAsStream( propFileCp );
         if( is != null ) {
            try { props.load( is ); } finally { is.close(); }
         }
      } catch( IOException e ) {/* nothing to do */}

      // Properties aus Datei per Pfad einlesen:
      if( props.isEmpty() ) {
         try( InputStream is = new FileInputStream( propFile ) ) {
            props.load( is );
         } catch( IOException e ) {/* nothing to do */}
      }
      return props;
   }

   /**
    * Extrahiere FTP-Parameter aus FTP-URI.
    * @param  uri FTP-URI, muss mit "ftp://" beginnen, z.B.: ftp://usr:pwd@host:21/path
    * @return Properties mit FTP-Parametern
    */
   public static Properties extractFtpParmsFromUri( String uri )
   {
      if( uri == null || uri.trim().length() <= FTP_PROT_URI.length() ||
                        !uri.trim().toLowerCase( Locale.GERMAN ).startsWith( FTP_PROT_URI ) ) {
         return null;
      }
      URI u;
      try {
         u = new URI( uri.trim() );
      } catch( URISyntaxException ex ) {
         throw new IllegalArgumentException( "Fehler: Ungueltige URI '" + uri + "': " + ex );
      }
      Properties p  = new Properties();
      String usrPwd = u.getUserInfo();
      String host   = u.getHost();
      int    port   = u.getPort();
      String path   = u.getPath();
      if( host == null || host.trim().length() == 0 ) { return null; }
      p.put( FTP_HOST_KEY, host );
      p.put( FTP_PORT_KEY, ( port > 0 ) ? (port + "") : "21" );
      if( path.length() > 0 ) { p.put( FTP_PATH_KEY, path ); }
      if( usrPwd != null && usrPwd.length() > 0 ) {
         String usr, pwd;
         int pos = usrPwd.indexOf( ':' );
         if( pos >= 0 ) {
            usr = usrPwd.substring( 0, pos ).trim();
            pwd = usrPwd.substring( pos + 1 ).trim();
         } else {
            usr = usrPwd.trim();
            pwd = "";
         }
         if( usr.length() > 0 ) { p.put( FTP_USR_KEY, usr ); }
         if( pwd.length() > 0 ) { p.put( FTP_PWD_KEY, pwd ); }
      }
      return p;
   }

   /**
    * Returniere Properties-Werte als Text-String, dabei wird das FTP-Passwort durch drei Punkte ersetzt.
    * @return Text-String der Properties-Werte, aber ohne FTP-Passwort
    */
   public String getPropsAsString()
   {
      return getPropsAsString( ftpOderFileProps );
   }

   /**
    * Returniere Properties-Werte als Text-String, dabei wird das FTP-Passwort durch drei Punkte ersetzt.
    * @param  p Properties
    * @return Text-String der Properties-Werte, aber ohne FTP-Passwort
    */
   public static String getPropsAsString( Properties p )
   {
      if( p == null ) { return null; }
      Properties propsOhnePwd = new Properties();
      propsOhnePwd.putAll( p );
      if( propsOhnePwd.get( FtpFileWrapper.FTP_PWD_KEY ) != null ) { propsOhnePwd.put( FtpFileWrapper.FTP_PWD_KEY, "..." ); }
      if( propsOhnePwd.get( FtpFileWrapper.FTP_URI_KEY ) != null ) {
         String s = propsOhnePwd.getProperty( FtpFileWrapper.FTP_URI_KEY );
         int pos2 = s.indexOf( '@' );
         int pos1 = ( pos2 > 0 ) ? s.lastIndexOf( ':', pos2 ) : -1;
         if( pos1 > 0 ) {
            propsOhnePwd.put( FtpFileWrapper.FTP_URI_KEY, s.substring( 0, pos1 + 1 ) + "..." + s.substring( pos2 ) );
         }
      }
      return propsOhnePwd.toString();
   }

   /**
    * Returniere Anzahl der Logins.
    * @return loginCount
    */
   public static long getLoginCount()
   {
      return LOGIN_COUNT.get();
   }

   /**
    * Zip: Einen InputStream in einen OutputStream zippen.
    * Wichtig: Auf beide Streams muss die close()-Methode aufgerufen werden.
    * @param instreamForZip  InputStream, der gezipped werden soll
    * @param outstreamZipped OutputStream, gezipped
    * @param origFilename    Originaldateiname (wird in der Zipdatei gespeichert)
    */
   public static void zipStream( InputStream instreamForZip, OutputStream outstreamZipped, String origFilename ) throws IOException
   {
      byte[] buf = new byte[65536];
      int len;
      ZipOutputStream zout = new ZipOutputStream( new BufferedOutputStream( outstreamZipped ) );
      zout.putNextEntry( new ZipEntry( origFilename ) );
      while( (len = instreamForZip.read( buf )) > 0 ) {
         zout.write( buf, 0, len );
      }
      zout.closeEntry();
      zout.finish();
      zout.flush();
   }

   /**
    * Unzip: Beim Entzippen von Dateien, die sehr viele Teildateien enthalten, kann es leicht passieren,
    * dass nicht genuegend freie Ports vorhanden sind.<br>
    * Pro zu entzippender Teildatei wird ein Port benoetigt, der fuer circa 2 Minuten blockiert bleibt.<br>
    * Deshalb bietet diese Methode die Moeglichkeit, eine Zeitverzoegerung pro zu entzippender Teildatei anzugeben.
    * @param  ffwrap                    FtpFileWrapper fuer den FTP-Zugang.
    * @param  remoteSourceZipFile       Zu entzippende Quell-Zipdatei.
    * @param  remoteDestDir             FTP-Zielverzeichnis.
    * @param  remoteFirstResultFilename falls null: der im Zip eingetragene Dateinamen, sonst erhaelt erste Datei diesen Namen
    * @param  minZeitProDateiMillisek   Minimale Zeit in Millisekunden pro entzippter Teildatei, z.B. 120 ms.
    * @return Anzahl entzippter Teildateien.
    */
   public static long unzipRemoteVerzoegert( FtpFileWrapper ffwrap, String remoteSourceZipFile, String remoteDestDir, String remoteFirstResultFilename, int minZeitProDateiMillisek ) throws IOException
   {
      Exception      ex = null;
      String         remoteResultFilename = null;
      long           anzahlEntries = 0;
      String         destDir       = ( remoteDestDir == null || remoteDestDir.trim().length() == 0 ) ? "" :
                                     ( remoteDestDir.trim().endsWith( "/" ) ) ? remoteDestDir.trim() : (remoteDestDir.trim() + "/");
      FtpFileWrapper ffwrapDest    = ffwrap.createSecondFtpFileWrapper();
      ffwrapDest.login();
      try {
         if( destDir != null && destDir.length() > 0 ) {
            ffwrapDest.changeWorkingDirectory( destDir );
         }
         try( InputStream instreamZipped = ffwrap.retrieveBinaryFile( remoteSourceZipFile ) ) {
            try( ZipInputStream zin = new ZipInputStream( new BufferedInputStream( instreamZipped ) ) ) {
               ZipEntry zipEntry;
               while( (zipEntry = zin.getNextEntry()) != null ) {
                  long startZeitNanosek = System.nanoTime();
                  remoteResultFilename = zipEntry.getName();
                  if( anzahlEntries == 0 && remoteFirstResultFilename != null && remoteFirstResultFilename.trim().length() > 0 ) {
                     remoteResultFilename = remoteFirstResultFilename;
                  }
                  if( remoteResultFilename != null && remoteResultFilename.startsWith( "/" ) && remoteResultFilename.length() > 1 ) {
                     remoteResultFilename = remoteResultFilename.substring( 1 );
                  }
                  try( BufferedOutputStream os = new BufferedOutputStream( ffwrapDest.storeFile( remoteResultFilename ) ) ) {
                     int size;
                     byte[] buffer = new byte[64 * 1024];
                     while( (size = zin.read( buffer, 0, buffer.length )) > 0 ) {
                        os.write( buffer, 0, size );
                     }
                  } catch( Exception e ) {
                     ex = e;
                  } finally {
                     ffwrapDest.completePendingCommand();
                  }
                  zin.closeEntry();
                  anzahlEntries++;
                  long verbrauchteZeitMillisek = (System.nanoTime() - startZeitNanosek) / NANOSEK_PRO_MILLISEK;
                  if( minZeitProDateiMillisek > verbrauchteZeitMillisek ) {
                     try { Thread.sleep( minZeitProDateiMillisek - verbrauchteZeitMillisek ); } catch( InterruptedException e ) {/*ok*/}
                  }
               }
            } catch( Exception e ) {
               if( ex == null ) { ex = e; }
            }
         } finally {
            ffwrap.completePendingCommand();
         }
      } finally {
         ffwrapDest.logout();
      }
      if( ex != null ) { throw new IOException( "Fehler beim Unzip von " + remoteSourceZipFile + ", letzter Zip-Entry " + remoteResultFilename + ",", ex ); }
      return anzahlEntries;
   }

   /**
    * Liste von Dateidaten rekursiv im Verzeichnis und in allen Unterverzeichnissen.
    * @param  startDir                 Startverzeichnis
    * @param  regexAllowedCompleteDirs Regex-Ausdruck fuer zu verwendende komplette Unterverzeichnispfade (also inkl. aller Unterverzeichnisse)
    * @param  ignoreSingleDirs         zu ignorierende Unterverzeichnisse, aber nicht komplette Unterverzeichnispfade, sondern einzelne Unterverzeichnisse
    * @return Liste von FtpFsFileData-Objekten (mit den Infos zu den Dateien)
    */
   public List<FtpFsFileData> listFtpFsFileDataRecursive( String startDir, String regexAllowedCompleteDirs, Set<String> ignoreSingleDirs )
   {
      loginIfNotLoggedInISE();
      List<FtpFsFileData> fds = listFtpFsFileData( startDir, false );
      if( fds == null ) { return null; }
      String prefixDir = ( startDir == null || startDir.trim().length() == 0 ) ? "" : (( startDir.trim().endsWith( "/" ) ) ? startDir.trim() : startDir.trim() + "/");
      Pattern regexAllowedCompleteDirsPattern = ( regexAllowedCompleteDirs != null && regexAllowedCompleteDirs.trim().length() > 0 ) ? Pattern.compile( regexAllowedCompleteDirs.trim() ) : null;
      List<FtpFsFileData> fdsResult = new ArrayList<FtpFsFileData>();
      for( FtpFsFileData fd : fds ) {
         if( fd.isFile ) {
            fdsResult.add( fd );
         } else if( fd.isDirectory ) {
            if( (regexAllowedCompleteDirsPattern == null || regexAllowedCompleteDirsPattern.matcher( prefixDir + fd.name ).matches()) &&
                (ignoreSingleDirs == null || !ignoreSingleDirs.contains( fd.name )) ) {
               fdsResult.addAll( listFtpFsFileDataRecursive( prefixDir + fd.name, regexAllowedCompleteDirs, ignoreSingleDirs ) );
            }
         }
      }
      return fdsResult;
   }

   /**
    * Liste Daten zu Dateien und Verzeichnissen.
    */
   public List<FtpFsFileData> listFtpFsFileData( String subDir )
   {
      return listFtpFsFileData( subDir, false );
   }

   /**
    * Liste Daten zu Dateien und Verzeichnissen.
    * Zu MLSD siehe {@link #listFtpFiles( String subDir, boolean withMLSD )}.
    */
   private List<FtpFsFileData> listFtpFsFileData( String subDir, boolean withMLSD )
   {
      loginIfNotLoggedInISE();
      List<FtpFsFileData> fds = new ArrayList<FtpFsFileData>();
      if( ftpOrFile ) {
         try {
            FTPFile[] ftpFiles = listFtpFiles( subDir, withMLSD );
            if( ftpFiles != null ) {
               for( FTPFile f : ftpFiles ) {
                  FtpFsFileData fd = new FtpFsFileData();
                  fd.timestamp   = f.getTimestamp();
                  fd.size        = f.getSize();
                  fd.name        = f.getName();
                  fd.isFile      = f.isFile();
                  fd.isDirectory = f.isDirectory();
                  fd.parentPath  = subDir;
                  fds.add( fd );
               }
            }
         } catch( IOException e ) {/* nothing to do*/}
      } else {
         File[] files = listFsFiles( subDir );
         if( files != null ) {
            for( File f : files ) {
               FtpFsFileData fd = new FtpFsFileData();
               Calendar cal = Calendar.getInstance();
               cal.setTimeInMillis( f.lastModified() );
               fd.timestamp   = cal;
               fd.size        = f.length();
               fd.name        = f.getName();
               fd.isFile      = f.isFile();
               fd.isDirectory = f.isDirectory();
               fd.parentPath  = subDir;
               fds.add( fd );
            }
         }
      }
      return fds;
   }

   /**
    * Liste Dateien und Verzeichnisse.
    * Nur Dateisystem, nicht FTP.
    */
   private File[] listFsFiles( String subDir )
   {
      String verz = ( subDir == null || subDir.trim().length() == 0 ) ? "" : subDir.trim();
      return (new File( getFsRootActPath( verz ) + verz )).listFiles();
   }

   /**
    * Liste Dateien und Verzeichnisse.
    * Nur FTP, nicht Dateisystem.
    * Ohne MLSD wird der Zeitstempel nicht immer korrekt ausgegeben, allerdings funktioniert MLSD nicht auf allen FTP-Servern.
    * Zu MLSD siehe {@link "http://tools.ietf.org/html/rfc3659"}.
    * @param subDir   Unterverzeichnis
    * @param withMLSD true falls MLSD verwendet werden soll
    * @return Array von FTPFile-Objekten
    */
   private FTPFile[] listFtpFiles( String subDir, boolean withMLSD ) throws IOException
   {
      FTPFile[] ftpFiles = null;
      if( subDir == null || subDir.trim().length() == 0 ) {
         ftpFiles = ( withMLSD ) ? ftpClient.mlistDir() : ftpClient.listFiles();
      } else {
         // Mit AS/400-FTP-Server ("OS/400", "V5R4M0") funktioniert "ftpClient.listFiles( subDir )" nicht:
         String oldWorkDir = getWorkingDirectory();
         try {
            if( changeWorkingDirectory( subDir ) ) {
               ftpFiles = ( withMLSD ) ? ftpClient.mlistDir() : ftpClient.listFiles();
            }
         } finally {
            if( !changeWorkingDirectory( oldWorkDir ) ) {
               throw new IOException( "Fehler: " + Arrays.asList( ftpClient.getReplyStrings() ) );
            }
         }
      }
      return ftpFiles;
   }

   /**
    * Ermittle Dateiinformationen zu einer Datei
    * @return Dateiinformationen
    */
   public FtpFsFileData getFtpFsFileData( String file )
   {
      FtpFsFileData fd = new FtpFsFileData();
      if( file == null || file.trim().length() == 0 ) {
         return fd;
      }
      loginIfNotLoggedInISE();
      if( ftpOrFile ) {
         File f = new File( file );
         String parentPath = f.getParent();
         String fileName   = f.getName();
         if( fileName == null || fileName.length() == 0 ) { return fd; }
         try {
            // Mit AS/400-FTP-Server ("OS/400", "V5R4M0") funktioniert "ftpClient.listFiles( subDir/fileName )" nicht:
            String oldWorkDir = getWorkingDirectory();
            try {
               if( changeWorkingDirectory( parentPath ) ) {
                  FTPFile[] ftpFiles = ftpClient.listFiles( fileName );
                  if( ftpFiles != null && ftpFiles.length == 1 ) {
                     fd.timestamp   = ftpFiles[0].getTimestamp();
                     fd.size        = ftpFiles[0].getSize();
                     fd.name        = ftpFiles[0].getName();
                     fd.isFile      = ftpFiles[0].isFile();
                     fd.isDirectory = ftpFiles[0].isDirectory();
                     fd.parentPath  = parentPath;
                  }
               }
            } finally {
               if( !changeWorkingDirectory( oldWorkDir ) ) {
                  throw new IOException( "Fehler: " + Arrays.asList( ftpClient.getReplyStrings() ) );
               }
            }
         } catch( IOException e ) {/* nothing to do*/}
      } else {
         File f = new File( getFsRootActPath( file ) + file );
         if( f.exists() ) {
            Calendar cal = Calendar.getInstance();
            cal.setTimeInMillis( f.lastModified() );
            fd.timestamp   = cal;
            fd.size        = f.length();
            fd.name        = f.getName();
            fd.isFile      = f.isFile();
            fd.isDirectory = f.isDirectory();
            fd.parentPath  = (new File( file )).getParent();
         }
      }
      return fd;
   }

   /** 
    * Ermittlung der FTP-Server-Software.
    * "ftpClient.getSystemType()" liefert beim FTP-Server "vsFTPd 2.2.2" nur: "UNIX Type: L8".
    * Deshalb wird zusaetzlich "ftpClient.getStatus()" verwendet.
   */
   public String ermittleFtpServerSoftware() throws IOException
   {
      if( !ftpOrFile ) {
         return "Fileserver";
      }
      String systType = ftpClient.getSystemType();
      if( systType != null ) {
         systType = systType.trim();
         String systTypeLowcase = systType.toLowerCase();
         if( systTypeLowcase.contains( "apache ftpserver" ) ||
             systTypeLowcase.contains( "proftpd" ) ||
             systTypeLowcase.contains( "vsftpd"  ) ||
             systTypeLowcase.contains( "os/400"  ) ) {
            if( systTypeLowcase.startsWith( "unix type:" ) && systType.length() > 12 ) {
               return systType.substring( 10 ).trim();
            }
            if( systType.length() > 2 ) {
               return systType;
            }
         }
      }
      String ftpServerName = null;
      String stat = ftpClient.getStatus();
      String[] ss = stat.split( "\n" );
      if( ss.length > 0 ) {
         String ss0 = ss[0].trim();
         if( ss0.startsWith( "211"       ) ) { ss0 = ss0.substring( 3 ).trim(); }
         if( ss0.startsWith( "-"         ) ) { ss0 = ss0.substring( 1 ).trim(); }
         if( ss0.startsWith( "Status of" ) ) { ss0 = ss0.substring( 9 ).trim(); }
         if( ss0.endsWith(   ":"         ) ) { ss0 = ss0.substring( 0, ss0.length() - 1 ).trim(); }
         if( !"FTP server status".equalsIgnoreCase( ss0 ) && !ss0.contains( "connected" ) ) {
            ftpServerName = ss0;
         }
         for( int i = 1; i < ss.length; i++ ) {
            if( ss[i].toLowerCase().contains( "vsftpd" ) ) {
               ftpServerName = ss[i].trim();
            }
         }
      }
      if( ftpServerName != null && ftpServerName.endsWith( "- secure, fast, stable" ) ) {
         ftpServerName = ftpServerName.substring( 0, ftpServerName.length() - 22 ).trim();
      }
      if( ftpServerName != null && ftpServerName.length() > 2 ) {
         return ftpServerName;
      }
      return systType;
   }

   /**
    * Nur im Fileserver-Betrieb:
    * @param  pathOrFile
    * @return Ermittle den voranzustellenden internen Pfad
    */
   private String getFsRootActPath( String pathOrFile )
   {
      if( pathOrFile != null && (pathOrFile.trim().startsWith( "/" ) || pathOrFile.trim().startsWith( "\\" )) ) {
         return fsRootDir;
      }
      return fsRootDir + fsActSubDir + File.separator;
   }

   /**
    * Fuehre login() aus, falls noch nicht eingeloggt.
    * @return true falls erfolgreich
    * @throws IllegalStateException statt IOException
    */
   private boolean loginIfNotLoggedInISE()
   {
      try {
         if( loginIfNotLoggedIn() ) { return loggedin; }
         throw new IllegalStateException( createLoginErrMsg() );
      } catch( IOException ex ) {
         throw new IllegalStateException( ex );
      }
   }

   /**
    * Ermittle, ob schon oder noch eingeloggt.
    * @return true falls eingeloggt
    */
   private boolean isLoggedIn()
   {
      if( !loggedin || !ftpOrFile ) {
         return loggedin;
      }
      if( ftpClient == null ) {
         return false;
      }
      try {
         ftpClient.printWorkingDirectory();
         return loggedin;
      } catch( IOException e ) {
         return false;
      }
   }

   /**
    * Fuehre login() aus, falls noch nicht eingeloggt.
    * @return true falls erfolgreich
    */
   private boolean loginIfNotLoggedIn() throws IOException
   {
      loggedin = isLoggedIn();
      if( loggedin ) {
         return loggedin;
      }
      try {
         String actSubDir = ftpActSubDir;
         if( login() ) {
            if( ftpOrFile && actSubDir != null && actSubDir.length() > 0 && !"/".equals( actSubDir ) ) {
               if( !ftpClient.changeWorkingDirectory( actSubDir ) ) {
                  throw new IOException( "FTP-Fehler mit FTP-Path '" + actSubDir + "': " + Arrays.asList( ftpClient.getReplyStrings() ) );
               }
            }
            return loggedin;
         }
         throw new IOException( createLoginErrMsg() );
      } catch( IOException ex ) {
         throw new IOException( createLoginErrMsg(), ex );
      }
   }

   /**
    * Returniere Login-Fehlermeldung inklusive der Properties-Werte als Text-String.
    * @return Login-Fehlermeldung
    */
   private String createLoginErrMsg()
   {
      return "Fehler: Kein FtpFileWrapper-Login moeglich, uebergebene Properties: " + getPropsAsString() + ".";
   }

   /**
    * Pruefe ob String nicht null ist, nicht leer ist und nicht nur aus Leerzeichen besteht.
    */
   private static boolean checkStringNichtLeer( String s )
   {
      return s != null && s.trim().length() > 0;
   }

   /**
    * finalize().
    */
   @Override
   protected void finalize() throws Throwable
   {
      logout();
      super.finalize();
   }

   public static class FtpFsFileData
   {
      public boolean  isFile;
      public boolean  isDirectory;
      public String   parentPath;
      public String   name;
      public long     size;
      public Calendar timestamp;
   }

   static class ConsoleProtocolCommandListener implements ProtocolCommandListener
   {
      @Override public void protocolCommandSent( ProtocolCommandEvent ev ) {
         System.out.print( "Sent:     " + ev.getMessage() );
      }
      @Override public void protocolReplyReceived( ProtocolCommandEvent ev ) {
         System.out.print( "Received: " + ev.getMessage() );
      }
   }
}
