package de.meinefirma.meinprojekt.ftp;

import java.io.*;
import org.apache.commons.net.ftp.FTPClient;
import org.apache.ftpserver.FtpServer;
import org.apache.ftpserver.ftplet.FtpException;
import org.junit.*;
import de.meinefirma.meinprojekt.ftp.FtpFileWrapper;

public class FtpFileWrapperIdleTimeoutTest
{
   private static final int    FTP_PORT               = 2121;
   private static final String FTP_HOST               = "localhost";
   private static final String FTP_HOME_DIR           = "target/FtpHome";
   private static final String READ_USER_NAME         = "ReadUserName";
   private static final String READ_USER_PWD          = "ReadUserPwd";
   private static final String WRITE_USER_NAME        = "WriteUserName";
   private static final String WRITE_USER_PWD         = "WriteUserPwd";
   private static final int    MAX_LOGINS             = 100;
   private static final int    MAX_IDLETIME_SEC_JUNIT = 1;
   private static final int    START_SEK_JUNIT        = MAX_IDLETIME_SEC_JUNIT + 1;
   private static final int    START_SEK_MAIN         = 605;
   private static final int    ANZAHL_VERSUCHE        = 10;
   private static FtpServer    ftpServer;

   /** Manueller Test mit externem FTP-Server:
    *  Ermittle die Laenge des Idle-Timeout */
   public static void main( String[] args )
   {
      String host           = ( args.length > 0 ) ? args[0] : "localhost";
      String port           = ( args.length > 1 ) ? args[1] : "4221";
      String usr            = ( args.length > 2 ) ? args[2] : "WriteUserName";
      String pwd            = ( args.length > 3 ) ? args[3] : "WriteUserPwd";
      String startSek       = ( args.length > 4 ) ? args[4] : "" + START_SEK_MAIN;
      String anzahlVersuche = ( args.length > 5 ) ? args[5] : "" + ANZAHL_VERSUCHE;
      testIdleTimeoutIntern( host, parseInt( port, 4221 ), usr, pwd, parseInt( startSek, START_SEK_MAIN ), parseInt( anzahlVersuche, ANZAHL_VERSUCHE ) );
   }

   @BeforeClass
   public static void startFtpServer() throws FtpException, IOException
   {
      ftpServer = FtpTestUtil.createFtpServer( FTP_PORT, FTP_HOME_DIR,
            READ_USER_NAME, READ_USER_PWD, WRITE_USER_NAME, WRITE_USER_PWD, null, MAX_LOGINS, MAX_IDLETIME_SEC_JUNIT );
      ftpServer.start();
   }

   @AfterClass
   public static void stoppFtpServer()
   {
      ftpServer.stop();
      ftpServer = null;
   }

   /** JUnit-Test:
    *  Ermittle die Laenge des Idle-Timeout */
   // Nur bei Bedarf aktivieren:
   // @Test
   public void testIdleTimeoutJUnit()
   {
      testIdleTimeoutIntern( FTP_HOST, FTP_PORT, WRITE_USER_NAME, WRITE_USER_PWD, START_SEK_JUNIT, ANZAHL_VERSUCHE );
   }

   /** JUnit-Test:
    *  Teste, ob nach Idle-Timeout und erneutem Login in das korrekte FTP-Verzeichnis gewechselt wurde */
   @Test
   public void testIdleTimeoutWorkingDirectoryJUnit() throws IOException, InterruptedException
   {
      testIdleTimeoutWorkingDirectoryIntern( "rootpath", "aktDir1", "aktDir2", "/aktDir1/aktDir2" );
      testIdleTimeoutWorkingDirectoryIntern( null, "aktDir1", "aktDir2", "/aktDir1/aktDir2" );
   }

   private static void testIdleTimeoutWorkingDirectoryIntern( String rootpath, String aktDir1, String aktDir2, String resultDir ) throws IOException, InterruptedException
   {
      StringBuilder sb = new StringBuilder();
      if( rootpath != null ) { sb.append( "\\" ).append( rootpath ); }
      if( aktDir1  != null ) { sb.append( "\\" ).append( aktDir1  ); }
      if( aktDir2  != null ) { sb.append( "\\" ).append( aktDir2  ); }
      (new File( FTP_HOME_DIR + sb.toString() + "\\lastDir" )).mkdirs();
      String uri = "ftp://" + WRITE_USER_NAME + ":" + WRITE_USER_PWD + "@" + FTP_HOST + ":" + FTP_PORT;
      if( rootpath != null ) { uri += "/" + rootpath; }
      FtpFileWrapper ffwrap = FtpFileWrapper.createFtpFileWrapperFromUri( uri );
      ffwrap.login();
      if( aktDir1  != null ) { ffwrap.changeWorkingDirectory( aktDir1 ); }
      if( aktDir2  != null ) { ffwrap.changeWorkingDirectory( aktDir2 ); }
      Assert.assertEquals( resultDir, ffwrap.getWorkingDirectory() );
      Thread.sleep( (MAX_IDLETIME_SEC_JUNIT + 1) * 1000 );
      Assert.assertEquals(  resultDir,  ffwrap.getWorkingDirectory() );
      Assert.assertEquals( "[lastDir]", ffwrap.listDirNames().toString() );
      ffwrap.logout();
   }

   private static void testIdleTimeoutIntern( String host, int port, String usr, String pwd, int startSek, int anzahlVersuche )
   {
      System.out.println( "host:port = " + host + ":" + port );
      int sec = startSek, secOld = startSek, secLastUnder = 0, secLastOver = 0;
      for( int i = 0; i < anzahlVersuche; i++ ) {
         System.out.println( "Versuch mit " + sec + " Sekunden ..." );
         secOld = sec;
         boolean timeOverIdleTimeout = checkTimeOverIdleTimeout( host, port, usr, pwd, sec );
         if( timeOverIdleTimeout ) {
            secLastOver = ( secLastOver == 0 ) ? sec : Math.min( sec, secLastOver );
            sec = (secLastOver + secLastUnder) / 2;
         } else {
            secLastUnder = Math.max( sec, secLastUnder );
            sec = ( secLastOver == 0 ) ? 2 * sec : (secLastOver + secLastUnder) / 2;
         }
         if( sec == 0 || secOld == sec || secLastUnder == secLastOver ) { break; }
      }
      if( secLastOver == 0 ) {
         System.out.println( "Der Idle-Timeout ist hoeher als " + sec + " Sekunden." );
      } else if( secLastUnder == secLastOver ) {
         System.out.println( "Der Idle-Timeout betraegt " + sec + " Sekunden." );
      } else {
         System.out.println( "Der Idle-Timeout liegt zwischen " + secLastUnder + " und " + secLastOver + " Sekunden." );
      }
   }

   static boolean checkTimeOverIdleTimeout( String host, int port, String usr, String pwd, int sec )
   {
      FTPClient ftpClient = new FTPClient();
      try {
         ftpClient.connect( host, port );
         ftpClient.login( usr, pwd );
         Thread.sleep( sec * 1000 );
         ftpClient.printWorkingDirectory();
         return false;
      } catch( Exception ex ) {
         return true;
      } finally {
         try { ftpClient.disconnect(); } catch( IOException e ) {/*ok*/}
      }
   }

   static int parseInt( String s, int returnIfNotParsable )
   {
      try { return Integer.parseInt( s.trim() ); } catch( Exception e ) { return returnIfNotParsable; }
   }

   static boolean parseBoolean( String s )
   {
      return s != null && s.trim().equalsIgnoreCase( "true" );
   }
}
