package de.meinefirma.meinprojekt.ftp;

import java.io.*;
import java.util.*;
import org.apache.ftpserver.FtpServer;
import org.apache.ftpserver.ftplet.FtpException;
import org.junit.*;
import de.meinefirma.meinprojekt.ftp.FtpFileWrapper.FtpFsFileData;

public class FtpFileWrapperTest
{
   private static final int    FTP_PORT        = 2121;
   private static final String FTP_HOST        = "localhost";
   private static final String FTP_PATH        = "ftp-path";
   private static final String FTP_HOME_DIR    = "target/FtpHome";
   private static final String READ_USER_NAME  = "ReadUserName";
   private static final String READ_USER_PWD   = "ReadUserPwd";
   private static final String WRITE_USER_NAME = "WriteUserName";
   private static final String WRITE_USER_PWD  = "WriteUserPwd";
   private static FtpServer ftpServer;

   @BeforeClass
   public static void startFtpServer() throws FtpException, IOException
   {
      ftpServer = FtpTestUtil.createFtpServer( FTP_PORT, FTP_HOME_DIR,
            READ_USER_NAME, READ_USER_PWD, WRITE_USER_NAME, WRITE_USER_PWD );
      ftpServer.start();
   }

   @AfterClass
   public static void stoppFtpServer()
   {
      // Um den FTP-Server von ausserhalb des Tests eine Zeit lang ueber
      // ftp://WriteUserName:WriteUserPwd@localhost:2121
      // anzusprechen, kann folgende Zeile aktiviert werden:
      // try { Thread.sleep( 55000 ); } catch( InterruptedException e ) {/*ok*/}
      ftpServer.stop();
      ftpServer = null;
   }

   /** Teste Fileserver-Variante, Properties-Datei nicht aus CLASSPATH sondern ueber Verzeichnisangabe */
   @Test
   public void testFileserverPropsAusVerzeichnis() throws IOException
   {
      FtpFileWrapper ffwrap = FtpFileWrapper.createFtpFileWrapperFromPropFile( "src/test/resources/FtpFileWrapper-FS.properties" );
      Assert.assertEquals( "{FsRoot=target/FileserverHome}", ffwrap.getPropsAsString() );
      testFtpFileWrapper( ffwrap, false );
   }

   /** Teste FTP-Variante, einzelne FTP-Parameter (keine URI), ohne FTP-Context-Path, Properties-Datei aus CLASSPATH */
   @Test
   public void testFtpPropsAusClasspath() throws IOException
   {
      FtpFileWrapper ffwrap = FtpFileWrapper.createFtpFileWrapperFromPropFile( "FtpFileWrapper-FTP.properties" );
      Assert.assertEquals( "{FtpHost=localhost, FtpPort=2121, FtpPwd=..., FtpUsr=WriteUserName}", ffwrap.getPropsAsString() );
      testFtpFileWrapper( ffwrap, true );
   }

   /** Teste FTP-Variante mit FTP-URI, ohne FTP-Context-Path, ohne Properties-Datei */
   @Test
   public void testFtpMitUriOhneContextPath() throws IOException
   {
      final String uri = "ftp://" + WRITE_USER_NAME + ":" + WRITE_USER_PWD + "@" + FTP_HOST + ":" + FTP_PORT;
      FtpFileWrapper ffwrap = FtpFileWrapper.createFtpFileWrapperFromUri( uri );
      Assert.assertEquals( "{FtpUri=ftp://WriteUserName:...@localhost:2121}", ffwrap.getPropsAsString() );
      testFtpFileWrapper( ffwrap, true );
   }

   /** Teste FTP-Variante mit FTP-URI, mit FTP-Context-Path, ohne Properties-Datei */
   @Test
   public void testFtpMitUriMitContextPath() throws IOException
   {
      final String uri = "ftp://" + WRITE_USER_NAME + ":" + WRITE_USER_PWD + "@" + FTP_HOST + ":" + FTP_PORT + "/" + FTP_PATH;
      FtpFileWrapper ffwrap = FtpFileWrapper.createFtpFileWrapperFromUri( uri );
      Assert.assertEquals( "{FtpUri=ftp://WriteUserName:...@localhost:2121/ftp-path}", ffwrap.getPropsAsString() );
      testFtpFileWrapper( ffwrap, true );
   }

   /** Interne gemeinsame FtpFileWrapper-Testmethode sowohl fuer FTP- als auch FS-Betrieb */
   protected static void testFtpFileWrapper( FtpFileWrapper ffwrap, boolean assureFtpOrFile ) throws IOException
   {
      String encoding              = "UTF-8";
      String meinText1             = "BlaBlupp";
      String meinText2             = "\u20AC"; // \u20AC = Euro-Zeichen
      String meinText3             = "Xyz";
      String meinText              = meinText1 + "\r\n" + meinText2;
      String remoteResultFileName1 = "MeineDatei1.txt";
      String remoteResultFileName2 = "MeineDatei2.txt";
      String remoteResultFileName3 = "MeineDatei3.txt";
      String remoteResultFilePath2 = "meindir2";
      String remoteResultFilePath3 = "meindir3";
      String remoteResultFile2     = remoteResultFilePath2 + "/" + remoteResultFileName2;
      String remoteResultFile3     = remoteResultFilePath3 + "/" + remoteResultFileName3;
      String remotePath4a          = "meindir4";
      String remotePath4b          = "meinsubdir4";
      String remoteResultFile5     = "d1/d2\\d3/r.txt";
      String remoteResultFile6     = "d1\\d2/d3\\f.txt";
      List<String> namesList;
      Assert.assertNotNull( "ffwrap null", ffwrap );

      // Test ohne expliziten Login:
      try {
         // FTP-Home-Path anlegen:
         if( assureFtpOrFile && new File( FTP_HOME_DIR ).exists() && !new File( FTP_HOME_DIR + "/" + FTP_PATH ).exists() ) {
            new File( FTP_HOME_DIR + "/" + FTP_PATH ).mkdirs();
         }

         // Erst nach Anlegen des FTP-Home-Paths aufrufen:
         Assert.assertEquals( "FtpOrFile", Boolean.valueOf( assureFtpOrFile ),
                                           Boolean.valueOf( ffwrap.getFtpOrFile() ) );

         // Schreiben von Remote-Datei ohne Unterverzeichnis:
         try( ByteArrayInputStream is1 = new ByteArrayInputStream( meinText.getBytes( encoding ) ) ) {
            Assert.assertTrue( "storeFile", ffwrap.storeFile( is1, remoteResultFileName1 ) );
         }
         namesList = ffwrap.listFileNames();
         Assert.assertNotNull( "listNames", namesList );
         Assert.assertTrue( "listNames", namesList.contains( remoteResultFileName1 ) );
         Assert.assertTrue( "exists", ffwrap.existsFile( remoteResultFileName1 ) );
         Assert.assertEquals("getLength", meinText.getBytes( encoding ).length, ffwrap.getLength( remoteResultFileName1 ) );

         // Schreiben von Remote-Datei mit Unterverzeichnis:
         try( ByteArrayInputStream is2 = new ByteArrayInputStream( meinText.getBytes( encoding ) ) ) {
            Assert.assertTrue( "storeFile", ffwrap.storeFile( is2, remoteResultFile2 ) );
         }
         namesList = ffwrap.listDirNames();
         Assert.assertNotNull( "listNames", namesList );
         Assert.assertTrue( remoteResultFilePath2, namesList.contains( remoteResultFilePath2 ) );
         namesList = ffwrap.listFileNames( remoteResultFilePath2 );
         Assert.assertNotNull( "listNames", namesList );
         Assert.assertTrue( remoteResultFileName2, namesList.contains( remoteResultFileName2 ) );
         Assert.assertTrue( "exists", ffwrap.existsFile( remoteResultFile2 ) );
         Assert.assertEquals("getLength", meinText.getBytes( encoding ).length, ffwrap.getLength( remoteResultFile2 ) );

         // Alternative Schreibmethode mit OutputStream:
         ffwrap.deleteFile( remoteResultFileName1 );
         try( BufferedWriter out = new BufferedWriter( new OutputStreamWriter( ffwrap.storeFile( remoteResultFileName1 ), encoding ) ) ) {
            out.write( meinText );
         }
         Assert.assertTrue( "completePendingCommand", ffwrap.completePendingCommand() );
         // Mit append:
         try( BufferedWriter out = new BufferedWriter( new OutputStreamWriter( ffwrap.storeFile( remoteResultFileName1, true ), encoding ) ) ) {
            out.write( "\r\n" + meinText3 );
         }
         Assert.assertTrue( "completePendingCommand", ffwrap.completePendingCommand() );

         // Zip:
         try( InputStream  isForZip = new ByteArrayInputStream( meinText.getBytes( encoding ) );
              OutputStream osZipped = ffwrap.storeFile( remoteResultFileName1 + ".zip" ) ) {
            FtpFileWrapper.zipStream( isForZip, osZipped, remoteResultFileName1 );
         }
         Assert.assertTrue( "completePendingCommand", ffwrap.completePendingCommand() );

         // Lesen von Datei ohne Unterverzeichnis:
         List<String> ss = ffwrap.retrieveTextLines( remoteResultFileName1, encoding, 100 );
         Assert.assertNotNull( "Text", ss );
         Assert.assertEquals( 3, ss.size() );
         Assert.assertEquals( meinText1, ss.get( 0 ) );
         Assert.assertEquals( meinText2, ss.get( 1 ) );
         Assert.assertEquals( meinText3, ss.get( 2 ) );

         // Lesen von Datei mit Unterverzeichnis:
         ss = ffwrap.retrieveTextLines( remoteResultFile2, encoding, 100 );
         Assert.assertNotNull( "Text", ss );
         Assert.assertEquals( 2, ss.size() );
         Assert.assertEquals( meinText1, ss.get( 0 ) );
         Assert.assertEquals( meinText2, ss.get( 1 ) );

         // changeWorkingDirectory():
         Assert.assertTrue( ffwrap.changeWorkingDirectory( remoteResultFilePath2 ) );
         namesList = ffwrap.listFileNames();
         Assert.assertTrue( "listNames", namesList.contains( remoteResultFileName2 ) );
         Assert.assertTrue( ffwrap.changeWorkingDirectory( "/" ) );
         namesList = ffwrap.listFileNames();
         Assert.assertTrue( "listNames", namesList.contains( remoteResultFileName1 ) );

         // moveRemoteFile() nach changeWorkingDirectory():
         Assert.assertTrue( ffwrap.changeWorkingDirectory( remoteResultFilePath2 ) );
         ffwrap.deleteFile( "/" + remoteResultFile3 );
         Assert.assertTrue( ffwrap.moveRemoteFile( remoteResultFileName2, "/" + remoteResultFile3 ) );
         Assert.assertTrue( ffwrap.changeWorkingDirectory( "/" ) );
         namesList = ffwrap.listDirNames();
         Assert.assertTrue( "moveRemoteFile(), Path", namesList.contains( remoteResultFilePath3 ) );
         namesList = ffwrap.listFileNames( remoteResultFilePath3 );
         Assert.assertTrue( "moveRemoteFile(), File", namesList.contains( remoteResultFileName3 ) );
         namesList = ffwrap.listFileNames( remoteResultFilePath2 );
         Assert.assertFalse( "moveRemoteFile(), Delete", namesList.contains( remoteResultFileName2 ) );

         // makeDirectory():
         ffwrap.makeDirectory( remotePath4a + "\\" + remotePath4b );
         namesList = ffwrap.listDirNames( remotePath4a );
         Assert.assertTrue( "makeDirectory()", namesList.contains( remotePath4b ) );

         // renameFile():
         try( ByteArrayInputStream is3 = new ByteArrayInputStream( "xy".getBytes() ) ) {
            Assert.assertTrue( "storeFile", ffwrap.storeFile( is3, remoteResultFile5 ) );
         }
         ffwrap.deleteFile( remoteResultFile6 );
         Assert.assertTrue( "renameFile", ffwrap.renameFile( remoteResultFile5, remoteResultFile6 ) );
         Assert.assertTrue( "existsFile", ffwrap.existsFile( remoteResultFile6 ) );

         // getFtpFsFileData():
         FtpFsFileData fd = ffwrap.getFtpFsFileData( "/" + remoteResultFile6 );
         Assert.assertEquals( "f.txt", fd.name );
         Assert.assertTrue(            fd.isFile );
         Assert.assertEquals(       2, fd.size );
         fd = ffwrap.getFtpFsFileData( "x.x" );
         Assert.assertEquals( null, fd.name );

         // Append: An Datei in Unterunterverzeichnis Text anhaengen:
         try( ByteArrayInputStream is4 = new ByteArrayInputStream( "z".getBytes() ) ) {
            Assert.assertTrue( "storeFile", ffwrap.storeFile( is4, remoteResultFile6, true ) );
         }
         ss = ffwrap.retrieveTextLines( remoteResultFile6, encoding, 100 );
         Assert.assertTrue( ss.size() == 1 );
         Assert.assertEquals( "xyz", ss.get( 0 ) );

         // listFtpFsFileDataRecursive():
         List<FtpFsFileData> fds;
         fds = ffwrap.listFtpFsFileDataRecursive( "/d1", "(/d1/d2|/d1/d2/.*)", null );
         Assert.assertEquals( "listFtpFsFileDataRecursive",                1, fds.size() );
         Assert.assertEquals( "listFtpFsFileDataRecursive", "/d1/d2/d3/f.txt", fds.get( 0 ).parentPath + "/" + fds.get( 0 ).name );
         Assert.assertEquals( "listFtpFsFileDataRecursive",                3, fds.get( 0 ).size );
         fds = ffwrap.listFtpFsFileDataRecursive( "d1", null, null );
         Assert.assertEquals( "listFtpFsFileDataRecursive",                1, fds.size() );
         Assert.assertEquals( "listFtpFsFileDataRecursive", "d1/d2/d3/f.txt", fds.get( 0 ).parentPath + "/" + fds.get( 0 ).name );
         Assert.assertEquals( "listFtpFsFileDataRecursive",                3, fds.get( 0 ).size );
         fds = ffwrap.listFtpFsFileDataRecursive( "d1", "(d1/d2|d1/d2/d3)", new HashSet<String>( Arrays.asList( new String[] { "xx" } ) ) );
         Assert.assertEquals( "listFtpFsFileDataRecursive",                1, fds.size() );
         Assert.assertEquals( "listFtpFsFileDataRecursive", "d1/d2/d3/f.txt", fds.get( 0 ).parentPath + "/" + fds.get( 0 ).name );
         Assert.assertEquals( "listFtpFsFileDataRecursive",                3, fds.get( 0 ).size );
         fds = ffwrap.listFtpFsFileDataRecursive( "d1", "xx", null );
         Assert.assertEquals( "listFtpFsFileDataRecursive", 0, fds.size() );
         fds = ffwrap.listFtpFsFileDataRecursive( "d1", null, new HashSet<String>( Arrays.asList( new String[] { "d2" } ) ) );
         Assert.assertEquals( "listFtpFsFileDataRecursive", 0, fds.size() );

         // Remove directory:
         Assert.assertTrue(  ffwrap.existsDir( "d1" ) );
         Assert.assertTrue(  ffwrap.removeDirectory( "d1" ) );
         Assert.assertFalse( ffwrap.existsDir( "d1" ) );

         // ObjectStreams:
         testFtpFileWrapperMitObjectStream( ffwrap );

         // Monitoring-Funktionen:
         long[] fileCountAndSizeSum = ffwrap.getFileCountAndSizeSum( null, null );
         FtpFsFileData oldestFileData = ffwrap.getOldestFileData( null, null, null );
         Assert.assertTrue( "fileCountAndSizeSum[0]: '" + fileCountAndSizeSum[0] + "' >= 3",   fileCountAndSizeSum[0] >= 3 );
         Assert.assertTrue( "fileCountAndSizeSum[1]: '" + fileCountAndSizeSum[1] + "' >= 152", fileCountAndSizeSum[1] >= 152 );
         Assert.assertTrue( "OldestFileData.timestamp", oldestFileData.timestamp.getTimeInMillis() > 1300000000000L );
         Assert.assertNotNull( "OldestFileData.name", oldestFileData.name );

      } finally {
         ffwrap.logout();
      }
   }

   /** Interne gemeinsame FtpFileWrapper-Testmethode fuer ObjectStreams sowohl fuer FTP- als auch FS-Betrieb */
   protected static void testFtpFileWrapperMitObjectStream( FtpFileWrapper ffwrap ) throws IOException
   {
      MyClass myObj1 = new MyClass();
      myObj1.s = "Blablupp";
      myObj1.i = 42;

      OutputStream os = ffwrap.storeFile( "MyObj.dat" );
      Assert.assertNotNull( "OutputStream null", os );
      try( ObjectOutputStream out = new ObjectOutputStream( os ) ) {
         out.writeObject( myObj1 );
         out.flush();
      } finally {
         ffwrap.completePendingCommand();
      }

      MyClass myObj2 = new MyClass();
      InputStream is = ffwrap.retrieveBinaryFile( "MyObj.dat" );
      Assert.assertNotNull( "InputStream null", is );
      try( ObjectInputStream in = new ObjectInputStream( is ) ) {
         myObj2 = (MyClass) in.readObject();
      } catch( ClassNotFoundException ex ) {
         Assert.fail( "ClassNotFoundException: " + ex.getMessage() );
      } finally {
         ffwrap.completePendingCommand();
      }
      Assert.assertNotNull( myObj2 );
      Assert.assertEquals( "Blablupp", myObj2.s );
      Assert.assertEquals(         42, myObj2.i );
   }
}

class MyClass implements Serializable
{
   private static final long serialVersionUID = 1L;
   String s;
   int i;
}
