package de.meinefirma.meinprojekt.ftp;

import java.io.*;
import java.util.*;
import org.apache.ftpserver.FtpServer;
import org.apache.ftpserver.ftplet.FtpException;
import org.junit.*;

/**
 * Test mit vielen parallelen konkurrierenden FTP-Zugriffen.<br>
 *<br>
 * Wenn diese Testklasse als JUnit-Test ausgefuehrt wird, wird ein temporaerer Test-FTP-Server gestartet.<br>
 * Alternativ kann mit der main()-Methode ein externer FTP-Server angesprochen werden (z.B. "localhost:4221").<br>
 *<br>
 * Achtung: Dieser Test kann nur wenige Male direkt hintereinander wiederholt ausgefuehrt werden!<br>
 * Bei zu haeufigen FTP-Logins und zu vielen FTP-Aktionen gibt es entweder eine Blockade oder die Exception:<br>
 *    java.net.BindException: Address already in use: connect <br>
 * Falls ein externer Apache-FtpServer verwendet wird, steht dann in dessen ftpd.log-Logfile:<br>
 *    java.io.IOException: Eine vorhandene Verbindung wurde vom Remotehost geschlossen <br>
 * Der Grund ist, dass nur begrenzt viele temporaer dynamisch zuordnungsbare Ports zur Verfuegung stehen.
 * Portnummern liegen im Bereich von 0 bis 65535 (dabei ist der Bereich bis 1024 fuer die "Well Known Ports" durch die IANA reserviert).
 * Die verschiedenen Betriebssysteme nutzen aber nur einen Teil des Bereichs:<br>
 * Z.B. Windows XP nutzt defaultmaessig nur den Bereich von 1025 bis 5000 (einstellbar in der Registry mit MaxUserPort unter
 * HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Services\Tcpip\Parameters).<br>
 * Unter Windows 7 ist normalerweise ein groesserer Bereich eingestellt, z.B. von 49152 bis 65535.
 * Dies kann ermittelt werden ueber:<br>
 *    netsh interface ipv4 show dynamicportrange tcp <br>
 * und eingestellt werden ueber:<br>
 *    netsh int ipv4 set dynamicport tcp start=49152 num=16384 <br>
 * Zu beachten ist, dass benutzte Ports nach der Verwendung noch einige Zeit vom Betriebssystem reserviert bleiben,
 * bei Windows normalerweise fuer 2 Minuten, beobachtbar ueber die Kommandos:<br>
 *    netstat -aon | findstr WARTEND <br>
 *    netstat -a | find /C "WARTEND" <br>
 * Weiterhin ist zu beachten, dass pro FTP-Verbindung mehrere Ports benoetigt werden: in diesem JUnit-Test ca. 12 Ports pro Task.<br>
 * Weitere Gruende fuer Probleme bei vielen FTP-Connections koennen sein:<br>
 * Falls Exception "java.net.SocketException: Too many files open": Siehe http://mina.apache.org/ftpserver/faq.html <br>
 * Konfiguration des FTP-Servers: Falls Apache FtpServer:
 * Eintraege zu max-threads und max-logins in der FTP-Config-XML-Datei
 * und zu ftpserver.user...maxloginnumber in der FTP-Users-Properties-Datei.
 */
public class FtpFileWrapperVielParallelTest
{
   private static final int    FTP_PORT        = 2121;
   private static final String FTP_HOST        = "localhost";
   private static final String FTP_HOME_DIR    = "target/FtpHome";
   private static final String READ_USER_NAME  = "ReadUserName";
   private static final String READ_USER_PWD   = "ReadUserPwd";
   private static final String WRITE_USER_NAME = "WriteUserName";
   private static final String WRITE_USER_PWD  = "WriteUserPwd";
   private static final String FS_HOME_DIR     = "target/FileserverHome";
   private static final int    MAX_LOGINS      = 100;
   private static final int    NUM_TASKS       = MAX_LOGINS / 3;
   private static final int    NUM_TREADS      = 16;
   private static FtpServer    ftpServer;

   /** Manueller Test mit externem FTP-Server */
   public static void main( String[] args ) throws IOException
   {
      String uri = ( args.length > 0 ) ? args[0] : "ftp://WriteUserName:WriteUserPwd@localhost:4221";
      testFtpFileWrapper( uri, "ftppath1", "ftppath2", true, NUM_TREADS, NUM_TASKS );
   }

   @BeforeClass
   public static void startFtpServer() throws FtpException, IOException
   {
      ftpServer = FtpTestUtil.createFtpServer( FTP_PORT, FTP_HOME_DIR,
            READ_USER_NAME, READ_USER_PWD, WRITE_USER_NAME, WRITE_USER_PWD, null, MAX_LOGINS );
      ftpServer.start();
   }

   @AfterClass
   public static void stoppFtpServer()
   {
      ftpServer.stop();
      ftpServer = null;
   }

   /** Teste Fileserver-Variante */
   @Test
   public void testFilesystem() throws IOException
   {
      testFtpFileWrapper( FS_HOME_DIR, false, NUM_TREADS, NUM_TASKS );
   }

   /** Teste FTP-Variante */
   @Test
   public void testFtpMitUri() throws IOException
   {
      testFtpFileWrapper( "ftp://" + WRITE_USER_NAME + ":" + WRITE_USER_PWD + "@" + FTP_HOST + ":" + FTP_PORT, true, NUM_TREADS, NUM_TASKS );
   }

   /** Interne gemeinsame FtpFileWrapper-Testmethode sowohl fuer FTP- als auch FS-Betrieb */
   protected static void testFtpFileWrapper( String uri, boolean isFtp, int threadCount, int taskCount ) throws IOException
   {
      testFtpFileWrapper( uri, "target/dst1", "target/dst2", isFtp, threadCount, taskCount );
   }

   /** Interne gemeinsame FtpFileWrapper-Testmethode sowohl fuer FTP- als auch FS-Betrieb */
   protected static void testFtpFileWrapper( String uri, String dstDir1, String dstDir2, boolean isFtp, int threadCount, int taskCount ) throws IOException
   {
      String charEnc  = "UTF-8";
      String meinText = "BlaBlupp\r\n\u20AC"; // \u20AC = Euro-Zeichen
      String srcDir   = "target/SrcDirConcurr";
      String homeDir  = ( isFtp ) ? FTP_HOME_DIR : FS_HOME_DIR;
      (new File( srcDir )).mkdirs();

      // Erzeuge die auszufuehrenden Task-Objekte und sammle in der taskList:
      Computable<TestFfwrapInput,String> testComputable = new TestFfwrapComputable();
      List<TaskDescription<TestFfwrapInput,String>> taskList = new ArrayList<TaskDescription<TestFfwrapInput,String>>();
      for( int i = 0; i < taskCount && i < MAX_LOGINS; i++ ) {
         String datei = i + ".txt";
         String srcDatei  = srcDir  + "/" + datei;
         String dstDatei1 = dstDir1 + "/" + datei;
         String dstDatei2 = dstDir2 + "/" + datei;
         // homeDir funktioniert nur im JUnit-Test und nicht mit main()-Methode:
         deleteFileIfExists( homeDir + "/" + dstDatei1 );
         deleteFileIfExists( homeDir + "/" + dstDatei2 );
         // Quelldateien nur anlegen, falls sie noch nicht existieren,
         // dadurch koennen auch groessere Dateien manuell in das "target/SrcDirConcurr"-Verzeichnis kopiert und verwendet werden:
         if( !(new File( srcDatei )).exists() ) {
            BufferedWriter out = new BufferedWriter( new OutputStreamWriter( new FileOutputStream( srcDatei ), charEnc ) );
            try { out.write( meinText ); } finally { out.close(); }
         }
         TestFfwrapInput inp = new TestFfwrapInput( uri, srcDatei, dstDatei1, dstDatei2 );
         taskList.add( new TaskDescription<TestFfwrapInput,String>( testComputable, inp ) );
      }
      int i = Math.max( uri.indexOf( '/' ), uri.indexOf( '@' ) );
      String s2 = ( isFtp && i > 0 ) ? uri.substring( ++i ) : uri;
      System.out.println( "Fuehre " + taskList.size() + " Tasks in " + threadCount + " Threads aus (" + s2 + "):" );
      long startZeit = System.nanoTime();

      // Fuehre die Tasks parallelisiert aus:
      (new TasksConcurrentExecutor<TestFfwrapInput,String>( taskList, threadCount, 3600 )).run();

      double dauer = ((System.nanoTime() - startZeit) / 1000000 / 1000.);
      boolean ok = true;
      for( TaskDescription<TestFfwrapInput,String> td : taskList ) {
         boolean okAktuell = td.result == null || td.result.equals( "null" );
         if( !okAktuell ) {
            System.out.println( "Fehler: arg=" + td.arg + ", result=" + td.result + ", Thread=" + Thread.currentThread().getName() );
         }
         ok &= okAktuell;
      }
      System.out.println( "Dauer: " + dauer + " s" );
      Assert.assertTrue( ok );
      FtpFileWrapper ffwrap = FtpFileWrapper.createFtpFileWrapperFromUri( uri );
      List<String> lst = ffwrap.listFileNames( dstDir1 );
      if( lst != null && lst.size() > 0 ) {
         System.out.println( "---- Warnung: '" + dstDir1 + "' entaelt '" + lst.size() + "' Dateien." );
      }
      lst = ffwrap.listFileNames( dstDir2 );
      if( lst != null && lst.size() < NUM_TASKS ) {
         System.out.println( "---- Warnung: '" + dstDir2 + "' entaelt nur '" + lst.size() + "' Dateien." );
      }
      ffwrap.logout();
   }

   static boolean deleteFileIfExists( String file )
   {
      File f = new File( file );
      return( f.exists() ) ? f.delete() : true;
   }
}

class TestFfwrapInput
{
   String uri;
   String srcDatei;
   String dstDatei1;
   String dstDatei2;

   public TestFfwrapInput( String uri, String srcDatei, String dstDatei1, String dstDatei2 ) {
      this.uri       = uri;
      this.srcDatei  = srcDatei;
      this.dstDatei1 = dstDatei1;
      this.dstDatei2 = dstDatei2;
   }
}

class TestFfwrapComputable implements Computable<TestFfwrapInput,String>
{
   @Override
   public String compute( TestFfwrapInput inp ) {
      FtpFileWrapper ffwrap = FtpFileWrapper.createFtpFileWrapperFromUri( inp.uri );
      Assert.assertNotNull( "ffwrap null", ffwrap );
      InputStream is1 = null;
      try {
         is1 = new BufferedInputStream( new FileInputStream( inp.srcDatei ) );
      } catch( FileNotFoundException ex ) {
         return ex.getMessage();
      }
      try {
         if( !ffwrap.storeFile( is1, inp.dstDatei1 ) ) {
            return "Fehler: ffwrap.storeFile( " + is1 + ", " + inp.dstDatei1 + " )";
         }
         if( !ffwrap.moveRemoteFile( inp.dstDatei1, inp.dstDatei2 ) ) {
            return "Fehler: ffwrap.moveRemoteFile( " + inp.dstDatei1 + ", " + inp.dstDatei2 + " )";
         }
         return null;
      } catch( IOException ex ) {
         String err = ex.getMessage();
         return ( err != null && err.trim().length() > 0 ) ? err : ex.toString();
      } finally {
         try { is1.close(); } catch( Exception e ) {/*ok*/}
         ffwrap.logout();
      }
   }
}
