package de.meinefirma.meinprojekt.ftp;

import java.util.List;
import java.util.concurrent.*;

/**
 * Parallele Abarbeitung vieler Tasks.
 * Aufruf ueber:
 * (new TasksConcurrentExecutor<InputType,ResultType>( taskList, threadCount, timeoutSeconds )).run();
 * Ergebnisse finden Sie in den Result-Objekten in den Tasks in der taskList.
 */
public class TasksConcurrentExecutor<A,R> implements Runnable
{
   private List<TaskDescription<A,R>> taskList;
   private int                        threadCount;
   private long                       timeoutSeconds;

   public TasksConcurrentExecutor( List<TaskDescription<A,R>> taskList, int threadCount, long timeoutSeconds )
   {
      this.taskList       = taskList;
      this.threadCount    = threadCount;
      this.timeoutSeconds = timeoutSeconds;
   }

   @Override public void run()
   {
      ExecutorService threadPool = Executors.newFixedThreadPool( threadCount );
      for( TaskDescription<A,R> td : taskList ) {
         threadPool.execute( new TaskRunnable<A,R>( td ) );
      }
      threadPool.shutdown();
      try {
         if( !threadPool.awaitTermination( timeoutSeconds, TimeUnit.SECONDS ) ) {
            threadPool.shutdownNow();
            throw new RuntimeException( "Timeout (" + timeoutSeconds + " Sekunden) ueberschritten." );
         }
      } catch( InterruptedException ex ) {
         throw new RuntimeException( ex );
      }
   }
}

class TaskRunnable<A,R> implements Runnable
{
   private TaskDescription<A,R> td;

   public TaskRunnable( TaskDescription<A,R> td )
   {
      this.td = td;
   }

   @Override public void run()
   {
      // System.out.println( Thread.currentThread().getName() + ": arg=" + td.arg );
      td.result = td.computable.compute( td.arg );
      // System.out.println( Thread.currentThread().getName() + ": result=" + td.result );
   }
}

/** Task-Beschreibung (Arbeitsmethode, Input-Argument und Ergebnis) */
class TaskDescription<A,R>
{
   Computable<A,R> computable;
   A               arg;
   R               result;

   public TaskDescription( Computable<A,R> computable, A arg )
   {
      this.computable = computable;
      this.arg        = arg;
   }
}

/** Interface fuer die Objekte mit der Task-Arbeitsmethode */
interface Computable<A,R>
{
   R compute( A arg );
}
